# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import torch
import torch.nn as nn

__all__ = ['ContrastiveLoss']


class ContrastiveLoss(nn.Module):
    """
    Contrastive loss function.

    References:
        https://github.com/delijati/pytorch-siamese/blob/master/contrastive.py

    LaTeX:
        Let D be the distance computed between the network layers or the direct
        distance output of the network.

        y is 0 if the pair should be labled as an imposter
        y is 1 if the pair should be labled as genuine

        ContrastiveLoss = ((y * D) ** 2 + ((1 - y) * max(m - D, 0) ** 2)) / 2

        $(y E)^2 + ((1 - y) max(m - E, 0)^2)$

    CommandLine:
        python -m clab.criterions ContrastiveLoss --show


    DisableExample:
        >>> # DISABLE_DOCTEST
        >>> from clab.criterions import *
        >>> import utool as ut
        >>> import numpy as np
        >>> vecs1, vecs2, label = testdata_siam_desc()
        >>> output = torch.nn.PairwiseDistance(p=2)(vecs1, vecs2)
        >>> self = ContrastiveLoss()
        >>> ut.exec_func_src(self.forward, globals())
        >>> func = self.forward
        >>> loss2x, dist = ut.exec_func_src(self.forward, globals(), globals(), keys=['loss2x', 'dist'])
        >>> ut.quit_if_noshow()
        >>> loss2x, dist, label = map(np.array, [loss2x, dist, label])
        >>> label = label.astype(np.bool)
        >>> dist0_l2 = dist[~label]
        >>> dist1_l2 = dist[label]
        >>> loss0 = loss2x[~label] / 2
        >>> loss1 = loss2x[label] / 2
        >>> # xdoc: +REQUIRES(--show)
        >>> import plottool as pt
        >>> pt.plot2(dist0_l2, loss0, 'x', color=pt.FALSE_RED, label='imposter_loss', y_label='loss')
        >>> pt.plot2(dist1_l2, loss1, 'x', color=pt.TRUE_BLUE, label='genuine_loss', y_label='loss')
        >>> pt.gca().set_xlabel('l2-dist')
        >>> pt.legend()
        >>> ut.show_if_requested()

    Example:
        >>> # DISABLE_DOCTEST
        >>> import torch
        >>> import netharn as nh
        >>> xpu = nh.XPU(None)
        >>> imgs1 = xpu.move(torch.rand(1, 3, 224, 244))
        >>> imgs2 = xpu.move(torch.rand(1, 3, 224, 244))
        >>> label = (xpu.move(torch.rand(3)) * 2).long()

        >>> model = SiameseLP(input_shape=imgs1.shape[1:])
        >>> output = model(imgs1, imgs2)
        >>> self = ContrastiveLoss(margin=10)
        >>> self.forward(output, label)
    """

    def __init__(self, weight=None, margin=1.0):
        # ut.super2(ContrastiveLoss, self).__init__()
        super(ContrastiveLoss, self).__init__()
        self.weight = weight
        self.margin = margin

        self.neg_label = 0
        self.pos_label = 1

    def forward(self, output, label):
        # Output should be a Bx1 vector representing the predicted
        # distance between each pair in a patch of image pairs
        dist = torch.squeeze(output)

        # Build indicator vectors labeling which pairs are pos and neg
        is_genuine = label.float()
        is_imposter = (1 - is_genuine)

        # Negate and clamp the distance for imposter pairs so these pairs are
        # encouraged to predict a distance larger than the margin
        hinge_neg_dist = torch.clamp(self.margin - dist, min=0.0)
        # Apply this loss only to negative examples
        loss_imposter = is_imposter * torch.pow(hinge_neg_dist, 2)

        # THe loss for positive examples is simply the distance because we wish
        # to encourage the network to predict values close to zero
        loss_genuine = is_genuine * torch.pow(dist, 2)

        # Weight each class if desired
        if self.weight is not None:
            loss_imposter *= self.weight[0]
            loss_genuine *= self.weight[1]

        # Sum the loss together (actually there is a 2x factor here)
        loss2x = loss_genuine + loss_imposter

        # Divide by 2 after summing for efficiency
        ave_loss = torch.sum(loss2x) / 2.0 / label.size()[0]
        loss = ave_loss
        return loss





def main():
    import xdoctest
    import ubelt as ub
    import sys

    modpath = ub.modname_to_modpath('netharn.examples')
    name_to_path = {}
    submods = list(xdoctest.static_analysis.package_modpaths(modpath))
    for submod in submods:
        modname = ub.augpath(submod, dpath='', ext='')
        if not modname.startswith('_'):
            name_to_path[modname] = submod

    print('name_to_path = {}'.format(ub.repr2(name_to_path, nl=1)))

    chosen = None
    for arg in sys.argv[1:2]:
        print('arg = {!r}'.format(arg))
        if arg in name_to_path:
            chosen = name_to_path[arg]
            break
    print('chosen = {!r}'.format(chosen))

    assert chosen is not None
    module = ub.import_module_from_path(chosen)
    print('module = {!r}'.format(module))
    module.main()

if __name__ == '__main__':
    """
    CommandLine:
        python -m netharn.examples
    """
    main()



import numpy as np
import torch
import torch.nn.functional as F
import itertools as it


def all_pairwise_distances(x, y=None, squared=False, approx=False):
    """
    Fast pairwise L2 squared distances between two sets of d-dimensional vectors

    Args:
        x (Tensor): an Nxd matrix
        y (Tensor, default=None): an optional Mxd matirx
        squared (bool, default=False): if True returns squared distances
        approx (bool, default=False): if True uses the quadratic distance approximation

    Returns:
        Tensor: dist: an NxM matrix where dist[i,j] is the square norm between
        x[i,:] and y[j,:] if y is not given then use 'y=x'.

        i.e. dist[i,j] = ||x[i,:]-y[j,:]||^2

    References:
        https://discuss.pytorch.org/t/efficient-distance-matrix-computation/9065

    Example:
        >>> from netharn.criterions.triplet import *
        >>> N, d = 5, 3
        >>> x = torch.rand(N, d)
        >>> dist = all_pairwise_distances(x)
        >>> assert dist.shape == (N, N)

        >>> a = x[None, :].expand(x.shape[0], -1, -1)
        >>> b = x[:, None].expand(-1, x.shape[0], -1)
        >>> real_dist = ((a - b) ** 2).sum(dim=2).sqrt_()
        >>> delta = (real_dist - dist.sqrt()).numpy()
        >>> assert delta.max() < 1e-5

    Example:
        >>> N, M, d = 5, 7, 3
        >>> x = torch.rand(N, d)
        >>> y = torch.rand(M, d)
        >>> dist = all_pairwise_distances(x, y)
        >>> assert dist.shape == (N, M)
    """
    if approx:
        return approx_pdist(x, y, squared=squared)
    else:
        return exact_pdist(x, y, squared=True)


def exact_pdist(x, y=None, squared=True):
    # Broadcast so all comparisons are NxN pairwise
    if y is None:
        y = x
    x_ = x[:, None, :]
    y_ = y[None, :, :]
    squared_dist = ((x_ - y_) ** 2).sum(dim=2)
    if squared:
        return squared_dist
    else:
        dist = squared_dist.sqrt_()
        return dist


def approx_pdist(x, y=None, squared=True):
    """
    This uses a quadratic expansion to approximate pairwise distances
    """
    x_norm = x.pow(2).sum(1).view(-1, 1)
    if y is None:
        y = x
        y_norm = x_norm.view(1, -1)
    else:
        y_norm = y.pow(2).sum(1).view(1, -1)
    yT = torch.transpose(y, 0, 1)
    xy = torch.mm(x, yT).mul_(-2.0)
    xy_norm = x_norm + y_norm
    squared_dist = xy_norm.add_(xy)
    squared_dist.clamp_(0, None)
    if squared:
        return squared_dist
    else:
        dist = squared_dist.sqrt_()
        return dist


def labels_to_adjacency_matrix(labels, symmetric=True, diagonal=True):
    """
    Construct an adjacency matrix of matching instances where `labels[i]` is
    the "name" or "identity" of the i-th item. The resulting matrix will have
    values adjm[i, j] == 1 if the i-th and j-th item have the same label and 0
    otherwise.

    Args:
        labels (ndarray): array of labels
        symmetric (bool, default=True): if False only the upper triangle of the
            matrix is populated.
        diagonal (bool, default=True): if False the diagonal is set to zero.

    Returns:
        ndarray: adjm : adjacency matrix

    Example:
        >>> labels = np.array([0, 0, 1, 1])
        >>> labels_to_adjacency_matrix(labels)
        array([[1, 1, 0, 0],
               [1, 1, 0, 0],
               [0, 0, 1, 1],
               [0, 0, 1, 1]], dtype=uint8)
        >>> labels_to_adjacency_matrix(labels, symmetric=False, diagonal=False)
        array([[0, 1, 0, 0],
               [0, 0, 0, 0],
               [0, 0, 0, 1],
               [0, 0, 0, 0]], dtype=uint8)
    """
    import netharn as nh
    n = len(labels)
    adjm = np.zeros((n, n), dtype=np.uint8)
    unique_labels, groupxs = nh.util.group_indices(labels)
    pos_idxs = [(i, j) for g in groupxs for (i, j) in it.combinations(sorted(g), 2)]
    pos_multi_idxs = tuple(zip(*pos_idxs))
    adjm[pos_multi_idxs] = 1
    if symmetric:
        adjm += adjm.T
    if diagonal:
        np.fill_diagonal(adjm, 1)
    return adjm


class TripletLoss(torch.nn.TripletMarginLoss):
    """
    Triplet loss with either hard or soft margin

    Example:
        >>> dvecs = torch.randn(21, 128)
        >>> labels = (torch.randn(len(dvecs)) * 4).long()
        >>> info = TripletLoss().mine_negatives(dvecs, labels)
        >>> pos_dists, neg_dists = info['pos_dists'], info['neg_dists']
        >>> self = TripletLoss(soft=1)
        >>> loss = self(pos_dists, neg_dists)
        >>> loss_s = TripletLoss(soft=1, reduction='none')(pos_dists, neg_dists)
        >>> loss_h = TripletLoss(margin=1, reduction='none')(pos_dists, neg_dists)

    Ignore:
        >>> import netharn as nh
        >>> xdata = torch.linspace(-10, 10)
        >>> ydata = {
        >>>     'soft_margin[0]': F.softplus(0 + xdata).numpy(),
        >>>     'soft_margin[1]': F.softplus(1 + xdata).numpy(),
        >>>     'soft_margin[4]': F.softplus(4 + xdata).numpy(),
        >>>     'hard_margin[0]': (0 + xdata).clamp_(0, None).numpy(),
        >>>     'hard_margin[1]': (1 + xdata).clamp_(0, None).numpy(),
        >>>     'hard_margin[4]': (4 + xdata).clamp_(0, None).numpy(),
        >>> }
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.multi_plot(xdata.numpy(), ydata, fnum=1)
    """

    def __init__(self, margin=1.0, eps=1e-6, reduction='mean', soft=False):
        super(TripletLoss, self).__init__(margin=margin, eps=eps,
                                          reduction=reduction)
        self.soft = soft

    def mine_negatives(self, dvecs, labels, num=1, mode='hardest', eps=1e-9):
        """
        triplets =
             are a selection of anchor, positive, and negative annots
             chosen to be the hardest for each annotation (with a valid
             pos and neg partner) in the batch.

        Args:
            dvecs (Tensor): descriptor vectors for each item
            labels (Tensor): id-label for each descriptor vector
            num (int, default=1): number of negatives per positive combination
            mode (str, default='hardest'): method for selecting negatives
            eps (float, default=1e9): distance threshold for near duplicates

        Returns:
            Dict: info: containing neg_dists, pos_dists, triples, and dist

        CommandLine:
            xdoctest -m netharn.criterions.triplet TripletLoss.mine_negatives:1 --profile

        Example:
            >>> from netharn.criterions.triplet import *
            >>> dvecs = torch.FloatTensor([
            ...     # Individual 1
            ...     [1.0, 0.0, 0.0, ],
            ...     [0.9, 0.1, 0.0, ],  # Looks like 2 [1]
            ...     # Individual 2
            ...     [0.0, 1.0, 0.0, ],
            ...     [0.0, 0.9, 0.1, ],  # Looks like 3 [3]
            ...     # Individual 3
            ...     [0.0, 0.0, 1.0, ],
            ...     [0.1, 0.0, 0.9, ],  # Looks like 1 [5]
            >>> ])
            >>> import itertools as it
            >>> labels = torch.LongTensor([0, 0, 1, 1, 2, 2])
            >>> num = 1
            >>> info = TripletLoss().mine_negatives(dvecs, labels, num)
            >>> print('info = {!r}'.format(info))
            >>> assert torch.all(info['pos_dists'] < info['neg_dists'])

        Example:
            >>> # xdoxctest: +SKIP
            >>> import itertools as it
            >>> p = 3
            >>> k = 10
            >>> d = p
            >>> mode = 'consistent'
            >>> mode = 'hardest'
            >>> for p, k in it.product(range(3, 13), range(2, 13)):
            >>>     d = p
            >>>     def make_individual_dvecs(i):
            >>>         vecs = torch.zeros((k, d))
            >>>         vecs[:, i] = torch.linspace(0.9, 1.0, k)
            >>>         return vecs
            >>>     dvecs = torch.cat([make_individual_dvecs(i) for i in range(p)], dim=0)
            >>>     labels = torch.LongTensor(np.hstack([[i] * k for i in range(p)]))
            >>>     num = 1
            >>>     info = TripletLoss().mine_negatives(dvecs, labels, num, mode=mode)
            >>>     if mode.startswith('hard'):
            >>>         assert torch.all(info['pos_dists'] < info['neg_dists'])
            >>>     base = k
            >>>     for a, p, n in info['triples']:
            >>>         x = a // base
            >>>         y = p // base
            >>>         z = n // base
            >>>         assert x == y, str([a, p, n])
            >>>         assert x != z, str([a, p, n])
        """
        import netharn as nh
        dist = all_pairwise_distances(dvecs, squared=True, approx=True)

        with torch.no_grad():
            labels_ = labels.numpy()

            symmetric = False
            pos_adjm = labels_to_adjacency_matrix(labels_, symmetric=symmetric,
                                                  diagonal=False)

            if symmetric:
                neg_adjm = 1 - pos_adjm
            else:
                neg_adjm = 1 - pos_adjm - pos_adjm.T
            np.fill_diagonal(neg_adjm, 0)

            # ignore near duplicates
            dist_ = dist.data.cpu().numpy()
            is_near_dup = np.where(dist_ < eps)
            neg_adjm[is_near_dup] = 0
            pos_adjm[is_near_dup] = 0

            # Filter out any anchor row that does not have both a positive and
            # a negative match
            flags = np.any(pos_adjm, axis=1) & np.any(neg_adjm, axis=1)

            anchors_idxs = np.where(flags)[0]
            pos_adjm_ = pos_adjm[flags].astype(bool)
            neg_adjm_ = neg_adjm[flags].astype(bool)

            if mode == 'hardest':
                # Order each anchors positives and negatives by increasing distance
                sortx_ = dist_[flags].argsort(axis=1)
                pos_cands_list = [x[m[x]] for x, m in zip(sortx_, pos_adjm_)]
                neg_cands_list = [x[n[x]] for x, n in zip(sortx_, neg_adjm_)]
                triples = []
                backup = []
                _iter = zip(anchors_idxs, pos_cands_list, neg_cands_list)
                for (anchor_idx, pos_cands, neg_cands) in _iter:
                    # Take `num` hardest negative pairs for each positive pair
                    num_ = min(len(neg_cands), len(pos_cands), num)
                    pos_idxs = pos_cands
                    neg_idxs = neg_cands[:num_]

                    anchor_dists = dist_[anchor_idx]

                    if True:
                        pos_dists = anchor_dists[pos_idxs]
                        neg_dists = anchor_dists[neg_idxs]

                        # Ignore any triple that satisfies the margin
                        # constraint
                        losses = pos_dists[:, None] - neg_dists[None, :] + self.margin
                        ilocs, jlocs = np.where(losses > 0)

                        if len(ilocs) > 0:
                            valid_pos_idxs = pos_idxs[ilocs].tolist()
                            valid_neg_idxs = neg_idxs[jlocs].tolist()

                            for pos_idx, neg_idx in zip(valid_pos_idxs, valid_neg_idxs):
                                triples.append((anchor_idx, pos_idx, neg_idx))
                        elif len(triples) == 0:
                            # Take everything because we might need a backup
                            for pos_idx, neg_idx in it.product(pos_idxs, neg_idxs):
                                backup.append((anchor_idx, pos_idx, neg_idx))
                    else:
                        for pos_idx, neg_idx in it.product(pos_idxs, neg_idxs):
                            # Only take items that will contribute positive loss
                            d_ap = anchor_dists[pos_idx]
                            d_an = anchor_dists[neg_idx]
                            loss = d_ap - d_an + self.margin
                            if loss > 0:
                                triples.append((anchor_idx, pos_idx, neg_idx))
                            elif len(triples) == 0:
                                backup.append((anchor_idx, pos_idx, neg_idx))

            elif mode == 'moderate':
                pos_cands_list = [np.where(m)[0].tolist() for m in pos_adjm_]
                neg_cands_list = [np.where(n)[0].tolist() for n in neg_adjm_]
                triples = []
                backup = []
                _iter = zip(anchors_idxs, pos_cands_list, neg_cands_list)
                for (anchor_idx, pos_cands, neg_cands) in _iter:
                    # Take `num` moderate negative pairs for each positive pair
                    # Only take items that will contribute positive loss
                    # but try not to take any that are too hard.
                    anchor_dists = dist_[anchor_idx]
                    neg_dists = anchor_dists[neg_cands]

                    for pos_idx in pos_cands:
                        pos_dist = anchor_dists[pos_idx]
                        losses = pos_dist - neg_dists + self.margin

                        # valid_negs = np.where((losses < margin) & (losses > 0))[0]
                        valid_negs = np.where(losses > 0)[0]
                        if len(valid_negs):
                            neg_idx = neg_cands[np.random.choice(valid_negs)]
                            triples.append((anchor_idx, pos_idx, neg_idx))
                        elif len(triples) == 0:
                            # We try to always return valid triples so, create
                            # a backup set in case we cant find any valid
                            # candidates
                            neg_idx = neg_cands[0]
                            backup.append((anchor_idx, pos_idx, neg_idx))

            elif mode == 'consistent':
                # Choose the same triples every time
                rng = nh.util.ensure_rng(0)
                pos_cands_list = [nh.util.shuffle(np.where(m)[0], rng=rng)
                                  for m in pos_adjm_]
                neg_cands_list = [nh.util.shuffle(np.where(n)[0], rng=rng)
                                  for n in neg_adjm_]
                triples = []
                _iter = zip(anchors_idxs, pos_cands_list, neg_cands_list)
                for (anchor_idx, pos_cands, neg_cands) in _iter:
                    num_ = min(len(neg_cands), len(pos_cands), num)
                    pos_idxs = pos_cands
                    for pos_idx in pos_idxs:
                        neg_idx = rng.choice(neg_cands)
                        triples.append((anchor_idx, pos_idx, neg_idx))
                rng.shuffle(triples)
            else:
                raise KeyError(mode)

            if len(triples) == 0:
                triples = backup
                if len(backup) == 0:
                    raise RuntimeError('unable to mine triples')

            triples = np.array(triples)
            A, P, N = triples.T

            if 0 and __debug__:
                if labels is not None:
                    for a, p, n in triples:
                        na_ = labels[a]
                        np_ = labels[p]
                        nn_ = labels[n]
                        assert na_ == np_
                        assert np_ != nn_

        # Note these distances are approximate distances, but they should be
        # good enough to backprop through (if not see alternate commented code).
        pos_dists = dist[A, P]
        neg_dists = dist[A, N]
        # pos_dists = (dvecs[A] - dvecs[P]).pow(2).sum(1)
        # neg_dists = (dvecs[A] - dvecs[N]).pow(2).sum(1)

        info = {
            'pos_dists': pos_dists,
            'neg_dists': neg_dists,
            'triples': triples,
            'dist': dist,
        }
        return info

    @classmethod
    def _hard_triples2(cls, dvecs, labels, margin=1):
        """
        Slow implementation of hard triples. Minimally modified from [1]

        References:
            ..[1] https://github.com/adambielski/siamese-triplet/blob/master/utils.py#L58
        """
        from itertools import combinations
        with torch.no_grad():
            embeddings = dvecs
            pdist = approx_pdist
            distance_matrix = pdist(embeddings)
            distance_matrix = distance_matrix.cpu()

            def semihard_negative(loss_values, margin=margin):
                semihard_negatives = np.where(np.logical_and(loss_values < margin, loss_values > 0))[0]
                return np.random.choice(semihard_negatives) if len(semihard_negatives) > 0 else None

            def hardest_negative(loss_values):
                hard_negative = np.argmax(loss_values)
                return hard_negative if loss_values[hard_negative] > 0 else None

            negative_selection_fn = hardest_negative

            labels_ = labels.cpu().data.numpy()
            triplets = []

            for label in set(labels_):
                label_mask = (labels_ == label)
                label_indices = np.where(label_mask)[0]
                if len(label_indices) < 2:
                    continue
                negative_indices = np.where(np.logical_not(label_mask))[0]
                anchor_positives = list(combinations(label_indices, 2))  # All anchor-positive pairs
                anchor_positives = np.array(anchor_positives)

                ap_distances = distance_matrix[anchor_positives[:, 0], anchor_positives[:, 1]]
                for anchor_positive, ap_distance in zip(anchor_positives, ap_distances):
                    loss_values = ap_distance - distance_matrix[torch.LongTensor(np.array([anchor_positive[0]])), torch.LongTensor(negative_indices)] + margin
                    loss_values = loss_values.data.cpu().numpy()
                    hard_negative = negative_selection_fn(loss_values)
                    if hard_negative is not None:
                        hard_negative = negative_indices[hard_negative]
                        triplets.append([anchor_positive[0], anchor_positive[1], hard_negative])

            if len(triplets) == 0:
                triplets.append([anchor_positive[0], anchor_positive[1], negative_indices[0]])

            triplets = np.array(triplets)

        ap_distances = (embeddings[triplets[:, 0]] - embeddings[triplets[:, 1]]).pow(2).sum(1)  # .pow(.5)
        an_distances = (embeddings[triplets[:, 0]] - embeddings[triplets[:, 2]]).pow(2).sum(1)  # .pow(.5)

        info = {
            'pos_dists': ap_distances,
            'neg_dists': an_distances,
            'triples': triplets,
            'dist': distance_matrix,
        }
        return info

    def _softmargin(self, pos_dists, neg_dists):
        x = pos_dists - neg_dists
        loss = F.softplus(x + self.margin)  # log(1 + exp(x))
        return loss

    def _hardmargin(self, pos_dists, neg_dists):
        x = pos_dists - neg_dists
        # loss = (self.margin + x).clamp_(0, None)  # [margin + x]_{+}
        loss = F.relu(x + self.margin)  # [margin + x]_{+}
        return loss

    def forward(self, pos_dists, neg_dists):
        """
        Args:
            pos_dists (Tensor) [A x 1]: distance between the anchor and a positive
            neg_dists (Tensor) [A x 1]: distance between the anchor and a negative

        Notes:
            soft_triplet_loss = (1 / len(triplets)) * sum(log(1 + exp(dist[a, p] - dist[a, n])) for a, p, n in triplets)

        Example:
            >>> from netharn.criterions.triplet import *
            >>> xbasis = ybasis = np.linspace(0, 5, 16)
            >>> pos_dists, neg_dists = map(torch.FloatTensor, np.meshgrid(xbasis, ybasis))
            >>> hard_loss = TripletLoss(reduction='none', soft=False).forward(pos_dists, neg_dists)
            >>> soft_loss = TripletLoss(reduction='none', soft=True).forward(pos_dists, neg_dists)
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> kwplot.plot_surface3d(pos_dists, neg_dists, hard_loss.numpy(), pnum=(1, 2, 1),
            >>>                       xlabel='d_pos', ylabel='d_neg', zlabel='hard-loss', contour=True, cmap='magma')
            >>> kwplot.plot_surface3d(pos_dists, neg_dists, soft_loss.numpy(), pnum=(1, 2, 2),
            >>>                       xlabel='d_pos', ylabel='d_neg', zlabel='hard-loss', contour=True, cmap='magma')
        """
        if self.soft:
            loss = self._softmargin(pos_dists, neg_dists)
        else:
            loss = self._hardmargin(pos_dists, neg_dists)
        if self.reduction == 'mean':
            loss = loss.mean()
        elif self.reduction == 'sum':
            loss = loss.sum()
        elif self.reduction != 'none':
            raise KeyError(self.reduction)
        return loss



"""
mkinit netharn.data
"""
# flake8: noqa

__DYNAMIC__ = False
if __DYNAMIC__:
    from mkinit import dynamic_init
    exec(dynamic_init(__name__))
else:
    # <AUTOGEN_INIT>
    from netharn.data import base
    from netharn.data import batch_samplers
    from netharn.data import coco_api
    from netharn.data import collate
    from netharn.data import mnist
    from netharn.data import toydata
    from netharn.data import transforms
    from netharn.data import voc

    from netharn.data.base import (DataMixin,)
    from netharn.data.batch_samplers import (MatchingSamplerPK,)
    from netharn.data.coco_api import (CocoDataset,)
    from netharn.data.collate import (CollateException, default_collate,
                                      list_collate, numpy_type_map,
                                      padded_collate,)
    from netharn.data.mnist import (MNIST,)
    from netharn.data.toydata import (ToyData1d, ToyData2d,)
    from netharn.data.voc import (VOCDataset,)

    __all__ = ['CocoDataset', 'CollateException', 'DataMixin', 'MNIST',
               'MatchingSamplerPK', 'ToyData1d', 'ToyData2d', 'VOCDataset', 'base',
               'batch_samplers', 'coco_api', 'collate', 'default_collate',
               'list_collate', 'mnist', 'numpy_type_map', 'padded_collate',
               'toydata', 'transforms', 'voc']



"""
FIXME 0 dimension tensors
"""
import torch.utils.data as torch_data
import torch
import ubelt as ub
import numpy as np  # NOQA
import re

# if six.PY2:
#     import collections
#     container_abcs = collections
# elif six.PY3:
#     import collections.abc
#     container_abcs = collections.abc
# string_classes = six.string_types
# int_classes = six.integer_types
from torch._six import container_abcs
from torch._six import string_classes, int_classes
default_collate = torch_data.dataloader.default_collate


# numpy_type_map = torch_data.dataloader.numpy_type_map  # moved in torch 1.1.0
numpy_type_map = {
    'float64': torch.DoubleTensor,
    'float32': torch.FloatTensor,
    'float16': torch.HalfTensor,
    'int64': torch.LongTensor,
    'int32': torch.IntTensor,
    'int16': torch.ShortTensor,
    'int8': torch.CharTensor,
    'uint8': torch.ByteTensor,
}


class CollateException(Exception):
    pass


_DEBUG = False


def _collate_else(batch, collate_func):
    """
    Handles recursion in the else case for these special collate functions

    This is duplicates all non-tensor cases from `torch_data.dataloader.default_collate`
    This also contains support for collating slices.
    """
    error_msg = "batch must contain tensors, numbers, dicts or lists; found {}"
    elem_type = type(batch[0])
    if elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
            and elem_type.__name__ != 'string_':
        elem = batch[0]
        if elem_type.__name__ == 'ndarray':
            # array of string classes and object
            if re.search('[SaUO]', elem.dtype.str) is not None:
                raise TypeError(error_msg.format(elem.dtype))

            return torch.stack([torch.from_numpy(b) for b in batch], 0)
        if elem.shape == ():  # scalars
            py_type = float if elem.dtype.name.startswith('float') else int
            return numpy_type_map[elem.dtype.name](list(map(py_type, batch)))
    elif isinstance(batch[0], slice):
        batch = default_collate([{
            'start': sl.start,
            'stop': sl.stop,
            'step': 1 if sl.step is None else sl.step
        } for sl in batch])
        return batch
    elif isinstance(batch[0], int_classes):
        return torch.LongTensor(batch)
    elif isinstance(batch[0], float):
        return torch.DoubleTensor(batch)
    elif isinstance(batch[0], string_classes):
        return batch
    elif isinstance(batch[0], container_abcs.Mapping):
        # Hack the mapping collation implementation to print error info
        if _DEBUG:
            collated = {}
            try:
                for key in batch[0]:
                    collated[key] = collate_func([d[key] for d in batch])
            except Exception:
                print('\n!!Error collating key = {!r}\n'.format(key))
                raise
            return collated
        else:
            return {key: collate_func([d[key] for d in batch]) for key in batch[0]}
    elif isinstance(batch[0], tuple) and hasattr(batch[0], '_fields'):  # namedtuple
        return type(batch[0])(*(default_collate(samples) for samples in zip(*batch)))
    elif isinstance(batch[0], container_abcs.Sequence):
        transposed = zip(*batch)
        return [collate_func(samples) for samples in transposed]
    else:
        raise TypeError((error_msg.format(type(batch[0]))))


def list_collate(inbatch):
    """
    Collates batches containing items with non-uniform data sizes

    Used for detection datasets with boxes.

    Args:
        inbatch: a list of items returned by __getitem__ for each item in the
            batch

    Example:
        >>> from netharn.data.collate import *
        >>> import torch
        >>> rng = np.random.RandomState(0)
        >>> inbatch = []
        >>> bsize = 4
        >>> for i in range(bsize):
        >>>     # add an image and some dummy bboxes to the batch
        >>>     img = torch.rand(3, 4, 4)  # dummy 4x4 image
        >>>     boxes = torch.LongTensor([[0, 0, 1, 1]] * i)
        >>>     item = (img, boxes)
        >>>     inbatch.append(item)
        >>> out_batch = list_collate(inbatch)
        >>> assert len(out_batch) == 2
        >>> batch_img, batch_boxes = out_batch
        >>> assert list(out_batch[0].shape) == [bsize, 3, 4, 4]
        >>> assert len(out_batch[1]) == bsize
        >>> assert len(out_batch[1][0]) == 0
        >>> assert len(out_batch[1][1]) == 1
        >>> assert len(out_batch[1][2]) == 2

    Example:
        >>> import torch
        >>> rng = np.random.RandomState(0)
        >>> inbatch = []
        >>> bsize = 4
        >>> for _ in range(bsize):
        >>>     # add an image and some dummy bboxes to the batch
        >>>     img = torch.rand(3, 8, 8)  # dummy 8x8 image
        >>>     boxes = torch.FloatTensor()
        >>>     item = (img, [boxes])
        >>>     inbatch.append(item)
        >>> out_batch = list_collate(inbatch)
        >>> assert len(out_batch) == 2
        >>> assert list(out_batch[0].shape) == [bsize, 3, 8, 8]
        >>> assert len(out_batch[1][0]) == bsize
    """
    try:
        if torch.is_tensor(inbatch[0]):
            num_items = [len(item) for item in inbatch]
            if ub.allsame(num_items):
                if len(num_items) == 0 or num_items[0] == 0:
                    batch = inbatch
                else:
                    batch = default_collate(inbatch)
            else:
                batch = inbatch
        else:
            batch = _collate_else(inbatch, list_collate)
    except Exception as ex:
        if not isinstance(ex, CollateException):
            raise CollateException(
                'Failed to collate inbatch={}. Reason: {!r}'.format(inbatch, ex))
        else:
            raise
    return batch


def padded_collate(inbatch, fill_value=-1):
    """
    Used for detection datasets with boxes.

    Example:
        >>> from netharn.data.collate import *
        >>> import torch
        >>> rng = np.random.RandomState(0)
        >>> inbatch = []
        >>> bsize = 7
        >>> for i in range(bsize):
        >>>     # add an image and some dummy bboxes to the batch
        >>>     img = torch.rand(3, 8, 8)  # dummy 8x8 image
        >>>     n = 11 if i == 3 else rng.randint(0, 11)
        >>>     boxes = torch.rand(n, 4)
        >>>     item = (img, boxes)
        >>>     inbatch.append(item)
        >>> out_batch = padded_collate(inbatch)
        >>> assert len(out_batch) == 2
        >>> assert list(out_batch[0].shape) == [bsize, 3, 8, 8]
        >>> assert list(out_batch[1].shape) == [bsize, 11, 4]

    Example:
        >>> import torch
        >>> rng = np.random.RandomState(0)
        >>> inbatch = []
        >>> bsize = 4
        >>> for _ in range(bsize):
        >>>     # add an image and some dummy bboxes to the batch
        >>>     img = torch.rand(3, 8, 8)  # dummy 8x8 image
        >>>     #boxes = torch.empty(0, 4)
        >>>     boxes = torch.FloatTensor()
        >>>     item = (img, [boxes])
        >>>     inbatch.append(item)
        >>> out_batch = padded_collate(inbatch)
        >>> assert len(out_batch) == 2
        >>> assert list(out_batch[0].shape) == [bsize, 3, 8, 8]
        >>> #assert list(out_batch[1][0].shape) == [bsize, 0, 4]
        >>> assert list(out_batch[1][0].shape) in [[0], []]  # torch .3 a .4

    Example:
        >>> inbatch = [torch.rand(4, 4), torch.rand(8, 4),
        >>>            torch.rand(0, 4), torch.rand(3, 4),
        >>>            torch.rand(0, 4), torch.rand(1, 4)]
        >>> out_batch = padded_collate(inbatch)
        >>> assert list(out_batch.shape) == [6, 8, 4]
    """
    try:
        if torch.is_tensor(inbatch[0]):
            num_items = [len(item) for item in inbatch]
            if ub.allsame(num_items):
                if len(num_items) == 0:
                    batch = torch.FloatTensor()
                elif num_items[0] == 0:
                    batch = torch.FloatTensor()
                else:
                    batch = default_collate(inbatch)
            else:
                max_size = max(num_items)
                real_tail_shape = None
                for item in inbatch:
                    if item.numel():
                        tail_shape = item.shape[1:]
                        if real_tail_shape is not None:
                            assert real_tail_shape == tail_shape
                        real_tail_shape = tail_shape

                padded_inbatch = []
                for item in inbatch:
                    n_extra = max_size - len(item)
                    if n_extra > 0:
                        shape = (n_extra,) + tuple(real_tail_shape)
                        if torch.__version__.startswith('0.3'):
                            extra = torch.Tensor(np.full(shape, fill_value=fill_value))
                        else:
                            extra = torch.full(shape, fill_value=fill_value,
                                               dtype=item.dtype)
                        padded_item = torch.cat([item, extra], dim=0)
                        padded_inbatch.append(padded_item)
                    else:
                        padded_inbatch.append(item)
                batch = inbatch
                batch = default_collate(padded_inbatch)
        else:
            batch = _collate_else(inbatch, padded_collate)
    except Exception as ex:
        if not isinstance(ex, CollateException):
            try:
                _debug_inbatch_shapes(inbatch)
            except Exception:
                pass
            raise CollateException(
                'Failed to collate inbatch={}. Reason: {!r}'.format(inbatch, ex))
        else:
            raise
    return batch


def _debug_inbatch_shapes(inbatch):
    import ubelt as ub
    print('len(inbatch) = {}'.format(len(inbatch)))
    extensions = ub.util_format.FormatterExtensions()

    @extensions.register((torch.Tensor, np.ndarray))
    def format_shape(data, **kwargs):
        return ub.repr2(dict(type=str(type(data)), shape=data.shape), nl=1, sv=1)

    print('inbatch = ' + ub.repr2(inbatch, extensions=extensions, nl=True))


if __name__ == '__main__':
    """
    CommandLine:
        python -m netharn.data.collate all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
The X Processing Unit --- an agnostic torch device.

An XPU is an abstracted (X) procesesing unit (PU) with a common API for running
torch operations on a CPU, GPU, or many GPUs.
"""
from __future__ import absolute_import, division, print_function
import ubelt as ub
import warnings
import torch
import six
import os
from netharn import util
from torch._six import container_abcs


__all__ = ['XPU']

# try:
# minimum memory (MB) needed for auto to resolve to GPU by default
NETHARN_MIN_MB = int(os.environ.get('NETHARN_MIN_MB', 6000))
# except Exception:
#     NETHARN_MIN_MB = 6000


# if torch.__version__.startswith('0.3'):
#     _TENSOR_TYPES = (torch._TensorBase, torch.autograd.Variable)
# else:
_TENSOR_TYPES = (torch.Tensor, torch.autograd.Variable)


class MountedModel(torch.nn.Module, util.ModuleMixin):
    """
    Abstraction of DataParallel and DataSerial
    """

    def receptive_field_for(self, input_field=None):
        import netharn as nh
        return nh.ReceptiveFieldFor(self.module)(input_field)


class DataParallel(torch.nn.DataParallel, MountedModel):
    """
    Hack to redefine DataParallel such that it shares a base with DataSerial
    """
    pass


class DataSerial(MountedModel):
    """
    Wraper to create consistent API with DataParallel
    """
    def __init__(self, module):
        super(DataSerial, self).__init__()
        self.module = module

    def forward(self, *inputs, **kwargs):
        return self.module.forward(*inputs, **kwargs)


class XPU(ub.NiceRepr):
    """
    A processing device or devices: either a CPU, GPU, or multiple GPUS.

    Args:
        item (None, int, or list): None for cpu, an int for a gpu, or a list of
            ints for multiple gpus.
    TODO:
        distributed processing

    CommandLine:
        python -m netharn.device XPU

    Example:
        >>> print(str(XPU(None)))
        CPU
        >>> print(str(XPU(0, check=False)))
        GPU(0)
        >>> print(str(XPU([1, 2, 3], check=False)))
        GPU(1*,2,3)
        >>> import pytest
        >>> with pytest.raises(IndexError):
        >>>     print(str(XPU([], check=False)))
    """
    def __init__(xpu, item=None, check=True):
        xpu._main_device_id = None
        xpu._device_ids = None
        xpu.mode = None

        # For context manager
        xpu._cuda_device = None

        if isinstance(item, torch.device):
            if item.type == 'cpu':
                item = None
            elif item.type == 'cuda':
                item = item.index or 0
            else:
                raise KeyError(item.type)
        elif isinstance(item, six.string_types):
            item = item.lower()
            if item in ['cpu', 'none']:
                item = None
            elif item.startswith(('gpu', 'cuda')):
                item = item.replace('gpu', '')
                item = item.replace('cuda', '')
                if item == '':
                    item = 0
                elif ',' in item:
                    item = list(map(int, item.split(',')))
                else:
                    item = int(item)
            else:
                raise KeyError('Unknown XPU string coding')

        if check:
            if not XPU.exists(item):
                if isinstance(item, int) and not torch.cuda.is_available():
                    raise ValueError('XPU {} does not exist. '
                                     'CUDA is not available'.format(item))
                else:
                    raise ValueError('XPU {} does not exist.'.format(item))

        if item is None:
            xpu._main_device_id = None
            xpu._device_ids = None
        elif isinstance(item, six.integer_types):
            item = int(item)
            xpu._main_device_id = item
            xpu._device_ids = [item]
        elif isinstance(item, (list, tuple)):
            xpu._device_ids = list(item)
            if len(xpu._device_ids) == 0:
                raise IndexError('empty device list')
            xpu._main_device_id = xpu._device_ids[0]
        else:
            raise TypeError(xpu)

        if xpu._main_device_id is None:
            xpu.mode = 'cpu'
        else:
            if xpu._device_ids and len(xpu._device_ids) > 1:
                xpu.mode = 'multi-gpu'
            else:
                xpu.mode = 'gpu'

        if xpu._main_device_id is not None:
            xpu._cuda_device = torch.cuda.device(xpu._main_device_id)

    @classmethod
    def from_auto(XPU, min_memory=NETHARN_MIN_MB):
        """
        Determines what a CPU/GPU device to use based on hardware usage.

        Args:
            min_memory (int): min memory needed in bytes to default to GPU.
                defaults to envvar NETHARN_MIN_MB or 6000
        """
        if torch.cuda.is_available():
            n_available = torch.cuda.device_count()
            gpu_num = find_unused_gpu(min_memory=min_memory)
            if gpu_num is None or gpu_num >= n_available:
                gpu_num = None
        else:
            gpu_num = None
        xpu = XPU(gpu_num)
        return xpu

    @classmethod
    def from_argv(XPU, check=True, **kwargs):
        """
        Determine what CPU/GPU device to use based on sys.argv

        CommandLine:
            python -m netharn.device XPU.from_argv --gpu=0,1

        Example:
            >>> xpu = XPU.from_argv()
            >>> print(xpu)
        """
        item = ub.argval('--xpu', default=ub.NoParam)
        if item is not ub.NoParam:
            xpu = XPU.coerce(item)
        else:
            anygpu = ub.argflag('--gpu')
            if anygpu:
                gpu_num = XPU.default_gpu()
            else:
                gpu_num = ub.argval('--gpu', default=None)
            if ub.argflag('--cpu'):
                xpu = XPU(None, check=check)
            elif gpu_num is None:
                xpu = XPU.from_auto(**kwargs)
            else:
                if gpu_num.lower() == 'none':
                    xpu = XPU(None)
                if isinstance(gpu_num, six.string_types) and ',' in gpu_num:
                    _device_ids = list(map(int, gpu_num.split(',')))
                    xpu = XPU(_device_ids, check=check)
                else:
                    xpu = XPU(int(gpu_num), check=check)
        return xpu

    @classmethod
    def from_data(XPU, item, check=True, **kwargs):
        """
        Creates an XPU to represent the processing device(s) a Module, Tensor,
        or Variable currently exists on.

        Example:
            >>> xpu = XPU.from_data(torch.randn(3))
            >>> assert not xpu.is_gpu()
            >>> if torch.cuda.is_available():
            >>>     xpu = XPU.from_data(torch.randn(3).to('cuda'))
            >>>     assert xpu.is_gpu()
            >>>     for i in range(torch.cuda.device_count()):
            >>>         xpu = XPU.from_data(torch.randn(3).to(i))
            >>>         assert xpu.is_gpu()
            >>>         assert xpu._main_device_id == i
        """
        if hasattr(item, 'device'):
            return XPU(item.device, check=check)
        if hasattr(item, 'is_cuda'):
            if item.is_cuda:
                return XPU(item.get_device(), check=check)
            else:
                return XPU(None, check=check)
        elif hasattr(item, 'state_dict'):
            devices = [item.device for item in item.state_dict().values()]
            _device_ids = set()
            for device in devices:
                if device.type == 'cuda':
                    index = device.index or 0
                    _device_ids.add(index)
                else:
                    _device_ids.add(None)
            try:
                _device_ids = sorted(_device_ids)
            except TypeError:
                raise Exception('cannot currently mix CPU and GPU')
            return XPU(_device_ids, check=check)
        else:
            raise TypeError(type(item))

    of = from_data  # alias

    @classmethod
    def coerce(XPU, item, check=True, **kwargs):
        """
        Converts objects of many different types into an XPU.

        Args:
            item : special string, int, list, or None

        Example:
            >>> assert XPU.coerce('0', check=False) == XPU(0, check=False)
            >>> assert XPU.coerce('0,1,2', check=False) == XPU([0, 1, 2], check=False)
            >>> assert XPU.coerce('2,3,4', check=False) == XPU([2, 3, 4], check=False)
            >>> assert XPU.coerce('gpus=2,3,4', check=False) == XPU([2, 3, 4], check=False)
            >>> assert XPU.coerce([0, 1], check=False) == XPU([0, 1], check=False)
            >>> assert XPU.coerce(torch.Tensor()) == XPU(None)
            >>> assert XPU.coerce(None) == XPU(None)
            >>> assert XPU.coerce('auto', check=False) is not None
        """
        if isinstance(item, dict):
            item = item['xpu']  # allow coercion from a configuration dict
        try:
            if item is None:
                return XPU(item, check=check)
            elif isinstance(item, XPU):
                return item
            elif isinstance(item, _TENSOR_TYPES):
                return XPU.from_data(item, check=check)
            elif isinstance(item, torch.nn.Module):
                return XPU.from_data(item, check=check)
            elif isinstance(item, six.integer_types):
                return XPU(int(item), check=check)
            elif isinstance(item, (list, tuple)):
                return XPU(item, check=check)
            elif isinstance(item, six.string_types):
                if item == 'auto':
                    return XPU.from_auto(**kwargs)
                elif item == 'argv':
                    return XPU.from_argv(check=check, **kwargs)
                if item == 'cpu' or item is None:
                    return XPU(None, check=check)
                elif item == 'cpu' or item is None:
                    return XPU(None, check=check)
                else:
                    item = item.lower()
                    item = item.replace('=', '')
                    item = item.replace('cpu', '')
                    item = item.replace('gpus', '')
                    item = item.replace('gpu', '')
                    item = item.replace('cuda', '')
                    if item == '':
                        if torch.cuda.is_available():
                            item = XPU.default_gpu()
                        else:
                            item = None
                    elif ',' in item:
                        item = list(map(int, item.split(',')))
                    elif item == 'none':
                        item = None
                    else:
                        item = int(item)
                    return XPU(item, check=check)
            else:
                ValueError
        except Exception as ex:
            raise ValueError(
                'cannot cast to XPU. item={!r}. Caused by: {!r}'.format(
                    item, ex))

    @classmethod
    def cast(xpu, item, check=True, **kwargs):
        """
        Deprecated, use XPU.coerce instead.
        """
        import warnings
        warnings.warn('XPU.cast is deprecated, use XPU.coerce instead',
                      DeprecationWarning)
        return xpu.coerce(item, check=check, **kwargs)

    def __eq__(xpu, other):
        """
        Example:
            >>> assert XPU([0], check=False) == XPU(0, check=False)
            >>> assert XPU('gpu', check=False) == XPU(0, check=False)
            >>> assert XPU([1], check=False) != XPU(0, check=False)
            >>> assert XPU('cpu', check=False) == XPU(None, check=False)
            >>> assert XPU([0, 1], check=False) == XPU([0, 1], check=False)
            >>> assert XPU([0, 1], check=False) != XPU([1, 0], check=False)
            >>> assert 'numpy' != XPU([1, 0], check=False)
        """
        try:
            return (xpu._main_device_id == other._main_device_id and
                    xpu._device_ids == other._device_ids)
        except AttributeError:
            return False

    @property
    def devices(xpu):
        """ A list of torch devices represented by this XPU """
        if xpu.is_gpu():
            return [torch.device(type='cuda', index=index)
                    for index in xpu._device_ids]
        else:
            return [torch.device(type='cpu')]

    @property
    def device(xpu):
        """ alias for main_device """
        return xpu.main_device

    @property
    def main_device(xpu):
        """
        The main torch device represented by this XPU

        Example:
            >>> xpu = XPU(None)
            >>> print(repr(xpu.main_device))
            device(type='cpu')
        """
        if xpu.is_gpu():
            return torch.device(type='cuda', index=xpu._main_device_id)
        else:
            return torch.device(type='cpu')

    @classmethod
    def exists(XPU, item):
        """
        Determins if GPU/CPU exists

        Args:
            item (int or None):
        """
        if item is None:
            return True
        elif isinstance(item, six.integer_types):
            if item < 0:
                raise ValueError('gpu num must be positive not {}'.format(item))
            return (torch.cuda.is_available() and
                    item < torch.cuda.device_count())
        elif isinstance(item, (tuple, list)):
            return all(XPU.exists(i) for i in item)
        else:
            raise TypeError(type(item))

    def memory(self):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:psutil)
            >>> from netharn.device import *
            >>> print(ub.repr2(XPU.coerce(None).memory()))
            {
                'available': ...,
                'total': ...,
                'used': ...,
            }
            >>> # xdoctest: +REQUIRES(--cuda)
            >>> print(ub.repr2(XPU.coerce(0).memory()))
            {
                'available': ...,
                'total': ...,
                'used': ...,
            }

        """
        gpus = gpu_info()
        info = {
            'available': 0,
            'total': 0,
            'used': 0,
        }
        if self._device_ids is None:
            try:
                import psutil
            except ImportError:
                import warnings
                warnings.warn('using XPU.memory on the CPU requires psutil')
                raise
            tup = psutil.virtual_memory()
            MB = 1 / 2 ** 20
            info['total'] += tup.total * MB
            info['used'] += tup.used * MB
            info['available'] += tup.available * MB
        else:
            for id in self._device_ids:
                info['total'] += gpus[id]['mem_total']
                info['used'] += gpus[id]['mem_used']
                info['available'] += gpus[id]['mem_avail']
        return info

    def __str__(xpu):
        return xpu.__nice__()

    def __enter__(xpu):
        if xpu._cuda_device:
            xpu._cuda_device.__enter__()
        return xpu

    def __exit__(xpu, ex_type, ex_value, tb):
        if xpu._cuda_device:
            return xpu._cuda_device.__exit__(ex_type, ex_value, tb)

    def __nice__(xpu):
        if xpu.is_gpu():
            if xpu._device_ids and len(xpu._device_ids) > 1:
                parts = [str(n) + '*' if n == xpu._main_device_id else str(n)
                         for n in xpu._device_ids]
                return 'GPU({})'.format(','.join(parts))
            else:
                return 'GPU({})'.format(xpu._main_device_id)
        else:
            return 'CPU'

    def __json__(xpu):
        """
        String encoding of an XPU

        CommandLine:
            xdoctest -m ~/code/netharn/netharn/device.py XPU.__json__

        Example:
            >>> print(XPU(None).__json__())
            CPU
            >>> print(XPU(0, check=False).__json__())
            GPU(0)
            >>> print(XPU([1, 2, 3], check=False).__json__())
            GPU(1*,2,3)
        """
        return str(xpu)

    def __int__(xpu):
        return xpu._main_device_id

    def number_of_devices(xpu):
        """ The number of underlying devices abstracted by this XPU """
        return 1 if not xpu._device_ids else len(xpu._device_ids)

    def is_cpu(xpu):
        return xpu._main_device_id is None

    def is_gpu(xpu):
        """ True if running in single or parallel gpu mode """
        return xpu._main_device_id is not None
        # return 'gpu' in xpu.mode

    @staticmethod
    def raw(model):
        """
        Unmounts the original core model if it is mounted.

        Args:
            model (torch.nn.Module): a model (potentially mounted)

        Returns:
            torch.nn.Module:
                if `model` is mounted returns `model.module`
                otherwise, returns `model`
        """
        if isinstance(model, (MountedModel, torch.nn.DataParallel)):
            # Unwrap the core model
            model = model.module
        return model

    def mount(xpu, model):
        """
        Like move, but only for models.
        Note that this works inplace for non-Tensor objects.

        Args:
            model (torch.nn.Module): the model to mount

        Returns:
            DataSerial | DataParallel :
                the model mounted on the XPU (which may be multiple GPUs)

        Example:
            >>> model = torch.nn.Conv2d(1, 1, 1)
            >>> xpu = XPU()
        """
        # Unwrap the core model if necessary
        model = xpu.raw(model)
        model = xpu.move(model)
        if xpu._device_ids and len(xpu._device_ids) > 1:
            model = DataParallel(model, device_ids=xpu._device_ids,
                                 output_device=xpu._main_device_id)
        else:
            model = DataSerial(model)
        return model

    def move(xpu, data, **kwargs):
        """
        Moves the model onto the primary GPU or CPU.

        If the data is nested in a container (e.g. a dict or list) then this
        funciton is applied recursively to all values in the container.

        Note:
            This works by calling the `.to` method, which works inplace for
            torch Modules, but is not implace for raw Tensors.

        Args:
            data (torch.Module | torch.Tensor | Collection):
                raw data or a collection containing raw data.
            **kwargs : forwarded to `data.cuda`

        Returns:
            torch.Tensor: the tensor with a dtype for this device

        Example:
            >>> data = torch.FloatTensor([0])
            >>> if torch.cuda.is_available():
            >>>     xpu = XPU.coerce('gpu')
            >>>     assert isinstance(xpu.move(data), torch.cuda.FloatTensor)
            >>> xpu = XPU.coerce('cpu')
            >>> assert isinstance(xpu.move(data), torch.FloatTensor)
            >>> assert isinstance(xpu.move([data])[0], torch.FloatTensor)
            >>> assert isinstance(xpu.move({0: data})[0], torch.FloatTensor)
            >>> assert isinstance(xpu.move({data}), set)
        """
        try:
            if xpu.is_gpu():
                return data.to(xpu._main_device_id, **kwargs)
            else:
                return data.to('cpu')
        except AttributeError:
            # Recursive move
            if isinstance(data, container_abcs.Mapping):
                cls = data.__class__
                return cls((k, xpu.move(v)) for k, v in data.items())
            elif isinstance(data, (container_abcs.Sequence, container_abcs.Set)):
                cls = data.__class__
                return cls(xpu.move(v) for v in data)
            else:
                raise TypeError('Unknown type {}'.format(type(data)))

    @classmethod
    def default_gpu(XPU):
        """
        Example:
            >>> print(XPU.default_gpu())
        """
        if torch.cuda.is_available():
            return torch.cuda.current_device()
        else:
            return None

    def set_as_default(xpu):
        """
        Sets this device as the default torch GPU

        Example:
            >>> import pytest
            >>> XPU(None).set_as_default()
            >>> if torch.cuda.is_available():
            >>>     XPU(0).set_as_default()
            >>>     assert torch.cuda.current_device() == 0
        """
        if xpu.is_gpu():
            torch.cuda.set_device(xpu._main_device_id)
        else:
            torch.cuda.set_device(-1)

    def load(xpu, fpath):
        """
        Loads data from a filepath onto this XPU

        Args:
            fpath (str or file): path to torch data file or file-like object

        Example:
            >>> from os.path import join
            >>> dpath = ub.ensure_app_cache_dir('netharn')
            >>> fpath = join(dpath, 'foo.pt')
            >>> cpu = XPU(None)
            >>> data = torch.FloatTensor([0])
            >>> torch.save(data, fpath)
            >>> loaded = cpu.load(fpath)
            >>> assert all(data == loaded)
        """
        # print('Loading data onto {} from {}'.format(xpu, fpath))
        try:
            return torch.load(fpath, map_location=xpu._map_location)
        except Exception:
            print('XPU={} Failed to load fpath={}'.format(xpu, fpath))
            raise

    def _map_location(xpu, storage, location):
        """
        Helper for `xpu.load` used when calling `torch.load`

        Args:
            storage (torch.Storage) : the initial deserialization of the
                storage of the data read by `torch.load`, residing on the CPU.
            location (str): tag identifiying the location the data being read
                by `torch.load` was originally saved from.

        Returns:
            torch.Storage : the storage
        """
        if xpu.is_gpu():
            return storage.cuda(xpu._main_device_id)
        else:
            return storage

    def synchronize(xpu):
        """
        Should be used when benchmarking performance of GPU implementaions
        """
        if xpu.is_gpu():
            torch.cuda.synchronize()


def find_unused_gpu(min_memory=0):
    """
    Finds GPU with the lowest memory usage by parsing output of nvidia-smi

    Args:
        min_memory (int): disregards GPUs with fewer than `min_memory` free MB

    Returns:
        int or None: gpu num if a match is found otherwise None

    CommandLine:
        python -c "from netharn import device; print(device.find_unused_gpu(300))"

    Example:
        >>> if torch.cuda.is_available():
        >>>     item = find_unused_gpu()
        >>>     assert item is None or isinstance(item, int)
    """
    gpus = gpu_info()
    if not gpus:
        return None

    # Order GPUs by most available memory
    # gpu_avail_mem = {n: -gpu['mem_avail'] for n, gpu in gpus.items()}

    # Order GPUs by fewest compute processes, and then by available memory
    gpu_avail_mem = {n: (gpu['num_compute_procs'], -gpu['mem_avail'])
                     for n, gpu in gpus.items()}
    ranked_order = ub.argsort(gpu_avail_mem)

    for gpu_num in ranked_order:
        gpu = gpus[gpu_num]
        if gpu['mem_avail'] >= min_memory:
            return gpu_num
    return None


def _query_nvidia_smi(mode, fields):
    """
    Runs nvidia smi in query mode

    Args:
        mode (str): the query cli flag to pass to nvidia-smi
        fields (List[str]): csv header fields to query

    Returns:
        List[Dict[str, str]]: parsed csv output
    """
    header = ','.join(fields)
    command = [
        'nvidia-smi',
        '--{}={}'.format(mode, header),
        '--format=csv,noheader'
    ]
    info = ub.cmd(command)
    if info['ret'] != 0:
        print(info['out'])
        print(info['err'])
        raise Exception('unable to call nvidia-smi: ret={}'.format(
            info['ret']))
    rows = []
    for line in info['out'].split('\n'):
        line = line.strip()
        if line:
            parts = [p.strip() for p in line.split(',')]
            row = ub.dzip(fields, parts)
            rows.append(row)
    return rows


def gpu_info(new_mode=True):
    """
    Run nvidia-smi and parse output

    Args:
        new_mode: internal argument that changes the underlying implementation

    Returns:
        OrderedDict: info about each GPU indexed by gpu number

    Note:
        Not gaurenteed to work if CUDA is not installed.

    Warnings:
        if nvidia-smi is not installed

    CommandLine:
        xdoctest -m netharn.device gpu_info --cuda

    Example:
        >>> # xdoctest: +REQUIRES(--cuda)
        >>> from netharn.device import gpu_info
        >>> gpus = gpu_info()
        >>> assert len(gpus) == torch.cuda.device_count()
        >>> # xdoctest: +IGNORE_WANT
        >>> print('gpus = {}'.format(ub.repr2(gpus, nl=4)))
        gpus = {
            0: {
                'gpu_uuid': 'GPU-348ebe36-252b-46fa-8a97-477ae331f6f4',
                'index': '0',
                'mem_avail': 10013.0,
                'mem_total': 11170.0,
                'mem_used': 1157.0,
                'memory.free': '10013 MiB',
                'memory.total': '11170 MiB',
                'memory.used': '1157 MiB',
                'name': 'GeForce GTX 1080 Ti',
                'num': 0,
                'num_compute_procs': 1,
                'procs': [
                    {
                        'gpu_num': 0,
                        'gpu_uuid': 'GPU-348ebe36-252b-46fa-8a97-477ae331f6f4',
                        'name': '/usr/bin/python',
                        'pid': '19912',
                        'type': 'C',
                        'used_memory': '567 MiB',
                    },
                ],
            },
        }


    """
    pass

    """
    Ignore:

        # official nvidia-smi python bindings
        pip install nvidia-ml-py

        import pynvml

        # TODO: make more efficient calls to nvidia-smi

        utilization.gpu
        utilization.memory
        compute_mode
        memory.total
        memory.used
        memory.free
        index
        name
        count

        nvidia-smi pmon --count 1

        nvidia-smi  -h
        nvidia-smi  --help-query-compute-apps
        nvidia-smi  --help-query-gpu

        nvidia-smi --help-query-accounted-apps
        nvidia-smi --help-query-supported-clocks
        nvidia-smi --help-query-retired-pages
        nvidia-smi --query-accounted-apps="pid" --format=csv

        nvidia-smi  --query-gpu="index,memory.total,memory.used,memory.free,count,name,gpu_uuid" --format=csv
        nvidia-smi  --query-compute-apps="pid,name,gpu_uuid,used_memory" --format=csv
        nvidia-smi  --query-accounted-apps="gpu_name,pid" --format=csv

        import timerit
        ti = timerit.Timerit(40, bestof=5, verbose=2)
        for timer in ti.reset('new1'):
            with timer:
                gpu_info(True)
        for timer in ti.reset('old'):
            with timer:
                gpu_info(False)
        for timer in ti.reset('xml'):
            with timer:
                gpu_info('xml')

        xdev.profile_now(gpu_info)('xml')

        for timer in ti.reset('cmd'):
            with timer:
                ub.cmd(['nvidia-smi', '--query', '--xml-format'])

        for timer in ti.reset('check_output'):
            with timer:
                import subprocess
                subprocess.check_output(['nvidia-smi', '--query', '--xml-format'])
    """
    if new_mode == 'xml':
        # Parse info out of the nvidia xml query
        # note, that even though this has less calls to nvidia-smi, there
        # is a lot more output, which makes it the slowest method especially
        # for multi-gpu systems
        import xml.etree.ElementTree as ET

        info = ub.cmd(['nvidia-smi', '--query', '--xml-format'])
        if info['ret'] != 0:
            print(info['out'])
            print(info['err'])
            warnings.warn('Problem running nvidia-smi: ret='.format(
                info['ret']))
            return None
        xml_string = info['out']
        root = ET.fromstring(xml_string)

        gpus = {}
        for gpu_elem in root.findall('gpu'):
            gpu = {}
            gpu['uuid'] = gpu_elem.find('uuid').text
            gpu['name'] = gpu_elem.find('product_name').text
            gpu['num'] = int(gpu_elem.find('minor_number').text)
            gpu['procs'] = [
                {item.tag: item.text for item in proc_elem}
                for proc_elem in gpu_elem.find('processes')
            ]

            for item in gpu_elem.find('fb_memory_usage'):
                gpu['memory.' + item.tag] = item.text

            gpu['mem_used'] = float(gpu['memory.used'].strip().replace('MiB', ''))
            gpu['mem_total'] = float(gpu['memory.total'].strip().replace('MiB', ''))
            gpu['mem_avail'] = gpu['mem_total'] - gpu['mem_used']
            gpus[gpu['num']] = gpu

            # Let each GPU know how many processes are currently using it
            num_compute_procs = 0
            num_graphics_procs = 0
            for proc in gpu['procs']:
                if proc['type'] == 'C':
                    num_compute_procs += 1
                elif proc['type'] == 'G':
                    num_graphics_procs += 1
                else:
                    raise NotImplementedError(proc['type'])
            gpu['num_compute_procs'] = num_compute_procs
            gpu['num_graphics_procs'] = num_graphics_procs
        return gpus

    if new_mode:
        # This is slightly more robust than the old mode, but it also makes
        # more than one call to nvidia-smi and cannot return information about
        # graphics processes.
        fields = ['index', 'memory.total', 'memory.used', 'memory.free',
                  'name', 'gpu_uuid']
        mode = 'query-gpu'
        try:
            gpu_rows = _query_nvidia_smi(mode, fields)
        except Exception as ex:
            warnings.warn('Problem running nvidia-smi: {!r}'.format(ex))
            return None

        fields = ['pid', 'name', 'gpu_uuid', 'used_memory']
        mode = 'query-compute-apps'
        proc_rows = _query_nvidia_smi(mode, fields)

        # Coerce into the old-style format for backwards compatibility
        gpus = {}
        for row in gpu_rows:
            gpu = row.copy()
            num = int(gpu['index'])
            gpu['num'] = num
            gpu['mem_used'] = float(gpu['memory.used'].strip().replace('MiB', ''))
            gpu['mem_total'] = float(gpu['memory.total'].strip().replace('MiB', ''))
            gpu['mem_avail'] = gpu['mem_total'] - gpu['mem_used']
            gpu['procs'] = []
            gpus[num] = gpu

        gpu_uuid_to_num = {g['gpu_uuid']: gpu['num'] for g in gpus.values()}

        for row in proc_rows:
            # Give each GPU info on which processes are using it
            proc = row.copy()
            proc['type'] = 'C'
            proc['gpu_num'] = gpu_uuid_to_num[proc['gpu_uuid']]
            num = proc['gpu_num']
            gpus[num]['procs'].append(proc)

        WITH_GPU_PROCS = False
        if WITH_GPU_PROCS:
            # Hacks in gpu-procs if enabled
            import re
            info = ub.cmd('nvidia-smi pmon -c 1')
            for line in info['out'].split('\n'):
                line = line.strip()
                if line and not line.startswith("#"):
                    parts = re.split(r'\s+', line, maxsplit=7)
                    if parts[1] != '-':
                        header = [
                            'gpu_num', 'pid', 'type', 'sm', 'mem', 'enc',
                            'dec', 'name']
                        proc = ub.dzip(header, parts)
                        proc['gpu_num'] = int(proc['gpu_num'])
                        if proc['type'] == 'G':
                            gpu = gpus[proc['gpu_num']]
                            gpu['procs'].append(proc)
                            proc['gpu_uuid'] = gpu['gpu_uuid']

        for gpu in gpus.values():
            # Let each GPU know how many processes are currently using it
            num_compute_procs = 0
            num_graphics_procs = 0
            for proc in gpu['procs']:
                if proc['type'] == 'C':
                    num_compute_procs += 1
                elif proc['type'] == 'G':
                    num_graphics_procs += 1
                else:
                    raise NotImplementedError(proc['type'])

            # NOTE calling nvidia-smi in query mode does not seem to have
            # support for getting info about graphics procs.
            gpu['num_compute_procs'] = num_compute_procs
            if WITH_GPU_PROCS:
                gpu['num_graphics_procs'] = num_graphics_procs

    else:
        # This is the original implementation of this function. It parses the
        # direct output of nvidia smi, it is prone to failure if the format of
        # this program's output ever changes.
        try:
            result = ub.cmd('nvidia-smi')
            if result['ret'] != 0:
                warnings.warn('Problem running nvidia-smi.')
                return None
        except Exception:
            warnings.warn('Could not run nvidia-smi.')
            return {}

        lines = result['out'].splitlines()

        gpu_lines = []
        proc_lines = []
        current = None

        state = '0_gpu_read'

        for line in lines:
            if current is None:
                # Signals the start of GPU info
                if line.startswith('|====='):
                    current = []
            else:
                if state == '0_gpu_read':
                    if len(line.strip()) == 0:
                        # End of GPU info
                        state = '1_proc_read'
                        current = None
                    elif line.startswith('+----'):
                        # Move to the next GPU
                        gpu_lines.append(current)
                        current = []
                    else:
                        current.append(line)
                elif state == '1_proc_read':
                    if line.startswith('+----'):
                        # Move to the next GPU
                        # End of proc info
                        state = 'terminate'
                        break
                    else:
                        proc_lines.append(line)
                else:
                    raise AssertionError(state)

        def parse_gpu_lines(lines):
            line1 = lines[0]
            line2 = lines[1]
            gpu = {}
            gpu['name'] = ' '.join(line1.split('|')[1].split()[1:-1])
            gpu['num'] = int(' '.join(line1.split('|')[1].split()[0]))

            mempart = line2.split('|')[2].strip()
            part1, part2 = mempart.split('/')
            gpu['mem_used'] = float(part1.strip().replace('MiB', ''))
            gpu['mem_total'] = float(part2.strip().replace('MiB', ''))
            gpu['mem_avail'] = gpu['mem_total'] - gpu['mem_used']
            return gpu

        def parse_proc_line(line):
            inner = '|'.join(line.split('|')[1:-1])
            if 'no running processes found' in inner.lower():
                # Handle "No running processes found" case in issue #2
                return None

            parts = [p.strip() for p in inner.split(' ')]
            parts = [p for p in parts if p]

            index = int(parts[0])
            pid = int(parts[1])
            proc_type = str(parts[2])
            proc_name = str(parts[3])
            used_mem = float(parts[4].replace('MiB', ''))

            proc = {
                'gpu_num': index,
                'pid': pid,
                'type': proc_type,
                'name': proc_name,
                'used_mem': used_mem,
            }
            return proc

        gpus = {}
        for num, lines in enumerate(gpu_lines):
            gpu = parse_gpu_lines(lines)
            assert num == gpu['num'], (
                'nums ({}, {}) do not agree. probably a parsing error'.format(num, gpu['num']))
            assert num not in gpus, (
                'Multiple GPUs labeled as num {}. Probably a parsing error'.format(num))
            gpus[num] = gpu
            gpus[num]['procs'] = []

        for line in proc_lines:
            # Give each GPU info on which processes are using it
            proc = parse_proc_line(line)
            if proc is not None:
                num = proc['gpu_num']
                gpus[num]['procs'].append(proc)

        for gpu in gpus.values():
            # Let each GPU know how many processes are currently using it
            num_compute_procs = 0
            num_graphics_procs = 0
            for proc in gpu['procs']:
                if proc['type'] == 'C':
                    num_compute_procs += 1
                elif proc['type'] == 'G':
                    num_graphics_procs += 1
                else:
                    raise NotImplementedError(proc['type'])
            gpu['num_compute_procs'] = num_compute_procs
            gpu['num_graphics_procs'] = num_graphics_procs

    return gpus


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m netharn.device all
        pytest ~/code/netharn/netharn/device.py
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
Implementation slightly modified from [1]_.


References:
    ..[1] https://github.com/lukemelas/EfficientNet-PyTorch
"""
import torch
from torch import nn
from torch.nn import functional as F
from netharn import layers

import re
import math
import collections
from functools import partial
from torch.utils import model_zoo


class Conv2dDynamicSamePadding(nn.Conv2d, layers.AnalyticModule):
    """ 2D Convolutions like TensorFlow, for a dynamic image size """

    def __init__(self, in_channels, out_channels, kernel_size, stride=1, dilation=1, groups=1, bias=True):
        super(Conv2dDynamicSamePadding, self).__init__(in_channels,
                                                       out_channels,
                                                       kernel_size, stride, 0,
                                                       dilation, groups, bias)
        self.stride = self.stride if len(self.stride) == 2 else [self.stride[0]] * 2

    def forward(self, x):
        ih, iw = x.size()[-2:]
        kh, kw = self.weight.size()[-2:]
        sh, sw = self.stride
        oh, ow = math.ceil(ih / sh), math.ceil(iw / sw)
        pad_h = max((oh - 1) * self.stride[0] + (kh - 1) * self.dilation[0] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[1] + (kw - 1) * self.dilation[1] + 1 - iw, 0)
        if pad_h > 0 or pad_w > 0:
            x = F.pad(x, [pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2])
        return F.conv2d(x, self.weight, self.bias, self.stride, self.padding, self.dilation, self.groups)

    @classmethod
    def forsize(cls, image_size=None):
        """ Chooses static padding if you have specified an image size, and dynamic padding otherwise.
            Static padding is necessary for ONNX exporting of models. """
        if image_size is None:
            return Conv2dDynamicSamePadding
        else:
            return partial(Conv2dStaticSamePadding, image_size=image_size)

    def _analytic_forward(self, inputs, _OutputFor, _Output, _Hidden,
                          **kwargs):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:ndsampler)
            >>> from netharn.models.efficientnet import *  # NOQA
            >>> import netharn as nh
            >>> kwargs = layers.AnalyticModule._analytic_shape_kw()
            >>> globals().update(kwargs)
            >>> inputs = (1, 3, 224, 224)
            >>> self = Conv2dDynamicSamePadding(2, 3, 5)
            >>> outputs = self.output_shape_for(inputs)
            >>> import ubelt as ub
            >>> print(nh.util.align(ub.repr2(outputs.hidden, nl=-1), ':'))
        """
        hidden = _Hidden()
        x = inputs
        ih, iw = _OutputFor.shape(x)[-2:]
        kh, kw = self.weight.size()[-2:]
        sh, sw = self.stride
        oh, ow = math.ceil(ih / sh), math.ceil(iw / sw)
        pad_h = max((oh - 1) * self.stride[0] + (kh - 1) * self.dilation[0] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[1] + (kw - 1) * self.dilation[1] + 1 - iw, 0)
        if pad_h > 0 or pad_w > 0:
            pad = [pad_w // 2, pad_w - pad_w // 2,
                   pad_h // 2, pad_h - pad_h // 2]
            x = hidden['dynamic_padding'] = _OutputFor(F.pad)(x, pad)

        weight = self.weight
        bias = self.bias is not None
        stride = self.stride
        padding = self.padding
        dilation = self.dilation
        groups = self.groups

        y = hidden['conv'] = _OutputFor(F.conv2d)(x, weight, bias, stride,
                                                  padding, dilation, groups)
        outputs = _Output.coerce(y, hidden)
        return outputs


class Conv2dStaticSamePadding(nn.Conv2d, layers.AnalyticModule):
    """ 2D Convolutions like TensorFlow, for a fixed image size"""

    def __init__(self, in_channels, out_channels, kernel_size, image_size=None, **kwargs):
        super(Conv2dStaticSamePadding, self).__init__(in_channels, out_channels, kernel_size, **kwargs)
        self.stride = self.stride if len(self.stride) == 2 else [self.stride[0]] * 2

        # Calculate padding based on image size and save it
        assert image_size is not None
        ih, iw = image_size if type(image_size) == list else [image_size, image_size]
        kh, kw = self.weight.size()[-2:]
        sh, sw = self.stride
        oh, ow = int(math.ceil(ih / sh)), int(math.ceil(iw / sw))
        pad_h = max((oh - 1) * self.stride[0] + (kh - 1) * self.dilation[0] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[1] + (kw - 1) * self.dilation[1] + 1 - iw, 0)
        self.image_size = image_size
        self._pad = (pad_h, pad_w)
        self._pad_w = pad_w
        self._pad_h = pad_w
        if pad_h > 0 or pad_w > 0:
            self.static_padding = nn.ZeroPad2d((pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2))
        else:
            self.static_padding = layers.Identity()

    def forward(self, x):
        x = self.static_padding(x)
        x = F.conv2d(x, self.weight, self.bias, self.stride, self.padding, self.dilation, self.groups)
        return x

    def _analytic_forward(self, inputs, _OutputFor, _Output, _Hidden,
                          **kwargs):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:ndsampler)
            >>> from netharn.models.efficientnet import *  # NOQA
            >>> import netharn as nh
            >>> kwargs = layers.AnalyticModule._analytic_shape_kw()
            >>> globals().update(kwargs)
            >>> inputs = (1, 3, 224, 224)
            >>> self = Conv2dStaticSamePadding(2, 3, 5, image_size=[512, 512])
            >>> outputs = self.output_shape_for(inputs)
            >>> import ubelt as ub
            >>> print(nh.util.align(ub.repr2(outputs.hidden, nl=-1), ':'))
        """
        hidden = _Hidden()
        x = inputs
        x = hidden['static_padding'] = _OutputFor(self.static_padding)(x)
        y = hidden['conv'] = _OutputFor(F.conv2d)(
            x, self.weight, self.bias is not None, self.stride, self.padding,
            self.dilation, self.groups)
        outputs = _Output.coerce(y, hidden)
        return outputs


##################
# Model definition
##################

class MBConvBlock(layers.AnalyticModule):
    """
    Mobile Inverted Residual Bottleneck Block

    Args:
        block_args (BlockArgs): see :class:`Details`
        global_params (GlobalParam): see :class:`Details`

    Attributes:
        has_se (bool): Whether the block contains a Squeeze and Excitation layer.
    """

    def __init__(self, block_args, global_params):
        super(MBConvBlock, self).__init__()
        self._block_args = block_args
        self._bn_mom = 1 - global_params.batch_norm_momentum
        self._bn_eps = global_params.batch_norm_epsilon
        self.has_se = (self._block_args.se_ratio is not None) and (0 < self._block_args.se_ratio <= 1)
        self.id_skip = block_args.id_skip  # skip connection and drop connect

        # Get static or dynamic convolution depending on image size
        Conv2d = Conv2dDynamicSamePadding.forsize(image_size=global_params.image_size)

        # Expansion phase
        inp = self._block_args.input_filters  # number of input channels
        oup = self._block_args.input_filters * self._block_args.expand_ratio  # number of output channels
        if self._block_args.expand_ratio != 1:
            self._expand_conv = Conv2d(in_channels=inp, out_channels=oup, kernel_size=1, bias=False)
            self._bn0 = nn.BatchNorm2d(num_features=oup, momentum=self._bn_mom, eps=self._bn_eps)

        # Depthwise convolution phase
        k = self._block_args.kernel_size
        s = self._block_args.stride
        # Note: it is important to set the weight decay to be very low for the
        # depthwise convolutions
        self._depthwise_conv = Conv2d(
            in_channels=oup, out_channels=oup, groups=oup,  # groups makes it depthwise
            kernel_size=k, stride=s, bias=False)
        self._bn1 = nn.BatchNorm2d(num_features=oup, momentum=self._bn_mom, eps=self._bn_eps)

        # Squeeze and Excitation layer, if desired
        if self.has_se:
            num_squeezed_channels = max(1, int(self._block_args.input_filters * self._block_args.se_ratio))
            self._se_reduce = Conv2d(in_channels=oup, out_channels=num_squeezed_channels, kernel_size=1)
            self._se_expand = Conv2d(in_channels=num_squeezed_channels, out_channels=oup, kernel_size=1)

        # Output phase
        final_oup = self._block_args.output_filters
        self._project_conv = Conv2d(in_channels=oup, out_channels=final_oup, kernel_size=1, bias=False)
        # Note that the bn2 layer before the residual add, should be
        # initailized with gamma=0
        self._bn2 = nn.BatchNorm2d(num_features=final_oup, momentum=self._bn_mom, eps=self._bn_eps)
        self._bn2._residual_bn = True
        noli = global_params.noli
        self._noli = layers.rectify_nonlinearity(noli, dim=2)

    def forward(self, inputs, drop_connect_rate=None):
        """
        :param inputs: input tensor
        :param drop_connect_rate: drop connect rate (float, between 0 and 1)
        :return: output of block
        """

        # Expansion and Depthwise Convolution
        x = inputs
        if self._block_args.expand_ratio != 1:
            x = self._noli(self._bn0(self._expand_conv(inputs)))
        x = self._noli(self._bn1(self._depthwise_conv(x)))

        # Squeeze and Excitation
        if self.has_se:
            x_squeezed = F.adaptive_avg_pool2d(x, 1)
            x_squeezed = self._se_expand(self._noli(self._se_reduce(x_squeezed)))
            x = torch.sigmoid(x_squeezed) * x

        x = self._bn2(self._project_conv(x))

        # Skip connection and drop connect
        input_filters, output_filters = self._block_args.input_filters, self._block_args.output_filters
        if self.id_skip and self._block_args.stride == 1 and input_filters == output_filters:
            if drop_connect_rate:
                x = self.drop_connect(x, p=drop_connect_rate)
            x = x + inputs  # skip connection
        return x

    @classmethod
    def demo(MBConvBlock):
        layer_block_args, global_params = Details.build_efficientnet_params()
        block_args = layer_block_args[0]
        self = MBConvBlock(block_args, global_params)
        return self

    def _analytic_forward(self, inputs, _OutputFor, _Output, _Hidden,
                          **kwargs):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:ndsampler)
            >>> from netharn.models.efficientnet import *  # NOQA
            >>> import netharn as nh
            >>> self = MBConvBlock.demo()
            >>> kwargs = self._analytic_shape_kw()
            >>> globals().update(kwargs)
            >>> input_shape = inputs = (1, 32, 224, 224)
            >>> outputs = self.output_shape_for(input_shape)
            >>> import ubelt as ub
            >>> print(nh.util.align(ub.repr2(outputs.hidden, nl=-1), ':'))
        """
        hidden = _Hidden()

        # Expansion and Depthwise Convolution
        x = inputs
        if self._block_args.expand_ratio != 1:
            x = hidden['expand_conv'] = _OutputFor(self._expand_conv)(inputs)
            x = hidden['_bn0'] = _OutputFor(self._bn0)(x)
            x = hidden['_noli0'] = _OutputFor(self._noli)(x)

        x = hidden['depthwise_conv'] = _OutputFor(self._depthwise_conv)(x)
        x = hidden['_bn1'] = _OutputFor(self._bn1)(x)
        x = hidden['_noli1'] = _OutputFor(self._noli)(x)

        # Squeeze and Excitation
        if self.has_se:
            x_squeezed = hidden['_se_pool'] = _OutputFor(F.adaptive_avg_pool2d)(x, 1)
            x_squeezed = hidden['_se_reduce'] =  _OutputFor(self._se_reduce)(x_squeezed)
            x_squeezed = hidden['_se_noli'] = _OutputFor(self._noli)(x_squeezed)
            x_squeezed = hidden['_se_expand'] = _OutputFor(self._se_expand)(x_squeezed)
            x_squeezed = hidden['_se_sigmoid'] = _OutputFor(torch.sigmoid)(x_squeezed)
            x = hidden['_se_mul'] = _OutputFor.mul(x_squeezed, x)

        x = hidden['_project'] = _OutputFor(self._project_conv)(x)
        x = hidden['_bn2'] = _OutputFor(self._bn2)(x)

        # Skip connection and drop connect
        input_filters, output_filters = self._block_args.input_filters, self._block_args.output_filters
        if self.id_skip and self._block_args.stride == 1 and input_filters == output_filters:
            drop_connect_rate = kwargs.get('drop_connect_rate', 0)
            if drop_connect_rate:
                try:
                    x = self.drop_connect(x, p=drop_connect_rate)
                except Exception:
                    pass
                hidden['drop_connect'] = x

            # skip connection
            x = hidden['skip'] = _OutputFor.add(x, inputs)
        outputs = _Output.coerce(x, hidden)
        return outputs

    def drop_connect(self, inputs, p):
        """ Drop connect. """
        if not self.training:
            return inputs
        batch_size = inputs.shape[0]
        keep_prob = 1 - p
        random_tensor = keep_prob
        random_tensor += torch.rand([batch_size, 1, 1, 1], dtype=inputs.dtype, device=inputs.device)
        binary_tensor = torch.floor(random_tensor)
        output = inputs / keep_prob * binary_tensor
        return output


class EfficientNet(layers.AnalyticModule):
    """
    An EfficientNet model. Most easily loaded with the .from_name or .from_pretrained methods

    Args:
        blocks_args (List[BlockArgs]): arguments for each block
        global_params (GlobalParams): shared between blocks

    Example:
        >>> # xdoctest: +REQUIRES(module:ndsampler)
        >>> model = EfficientNet.from_name('efficientnet-b0')
        >>> print(model.number_of_parameters())
        >>> print(model.get_image_size())
        5288548
        224

        >>> model = EfficientNet.from_name('efficientnet-b8')
        >>> print(model.number_of_parameters())
        >>> print(model.get_image_size())
        87413142
        672
    """

    def __init__(self, blocks_args=None, global_params=None):
        super(EfficientNet, self).__init__()
        assert isinstance(blocks_args, list), 'blocks_args should be a list'
        assert len(blocks_args) > 0, 'block args must be greater than 0'
        self._global_params = global_params
        self._blocks_args = blocks_args

        # Handle class specification
        import ndsampler
        import ubelt as ub
        classes = self._global_params.classes
        if classes is None:
            classes = self._global_params.num_classes
        self.classes = ndsampler.CategoryTree.coerce(classes)

        keys = self._global_params._fields
        vals = list(self._global_params)
        tmp = ub.dzip(keys, vals, cls=ub.odict)
        tmp['num_classes'] = len(self.classes)
        tmp['classes'] = self.classes.__json__()
        self._global_params = type(global_params)(**tmp)

        self.image_size = self._global_params._asdict()['image_size']

        # import ubelt as ub
        # print(ub.repr2(self._global_params._asdict(), nl=-4))
        # print(ub.repr2(self._global_params._asdict()))

        self._initkw = {
            'blocks_args': self._blocks_args,
            'global_params': self._global_params,
        }

        self.model_name = None

        # Get static or dynamic convolution depending on image size
        Conv2d = Conv2dDynamicSamePadding.forsize(image_size=global_params.image_size)

        # Batch norm parameters
        bn_mom = 1 - self._global_params.batch_norm_momentum
        bn_eps = self._global_params.batch_norm_epsilon

        # Stem
        in_channels = 3  # rgb
        out_channels = self.round_filters(32)  # number of output channels
        self._conv_stem = Conv2d(in_channels, out_channels, kernel_size=3, stride=2, bias=False)
        self._bn0 = nn.BatchNorm2d(num_features=out_channels, momentum=bn_mom, eps=bn_eps)

        multiplier = global_params.depth_coefficient

        def round_repeats(repeats, multiplier):
            """ Round number of filters based on depth multiplier. """
            if not multiplier:
                return repeats
            return int(math.ceil(multiplier * repeats))

        # Build blocks
        self._blocks = nn.ModuleList([])
        for block_args in self._blocks_args:

            # Update block input and output filters based on depth multiplier.
            block_args = block_args._replace(
                input_filters=self.round_filters(block_args.input_filters),
                output_filters=self.round_filters(block_args.output_filters),
                num_repeat=round_repeats(block_args.num_repeat, multiplier)
            )

            # The first block needs to take care of stride and filter size increase.
            self._blocks.append(MBConvBlock(block_args, self._global_params))
            if block_args.num_repeat > 1:
                block_args = block_args._replace(input_filters=block_args.output_filters, stride=1)
            for _ in range(block_args.num_repeat - 1):
                self._blocks.append(MBConvBlock(block_args, self._global_params))

        # Head
        in_channels = block_args.output_filters  # output of final block
        out_channels = self.round_filters(1280)
        self._conv_head = Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        self._bn1 = nn.BatchNorm2d(num_features=out_channels, momentum=bn_mom, eps=bn_eps)

        # Final linear layer
        self._avg_pooling = nn.AdaptiveAvgPool2d(1)
        self._dropout = nn.Dropout(self._global_params.dropout_rate)
        self._fc = nn.Linear(out_channels, self._global_params.num_classes)
        noli = global_params.noli
        self._noli = layers.rectify_nonlinearity(noli, dim=2)

    def round_filters(self, filters):
        """ Calculate and round number of filters based on depth multiplier. """
        multiplier = self._global_params.width_coefficient
        if not multiplier:
            return filters
        divisor = self._global_params.depth_divisor
        min_depth = self._global_params.min_depth
        filters *= multiplier
        min_depth = min_depth or divisor
        new_filters = max(min_depth, int(filters + divisor / 2) // divisor * divisor)
        if new_filters < 0.9 * filters:  # prevent rounding by more than 10%
            new_filters += divisor
        return int(new_filters)

    def extract_features(self, inputs):
        """
        Returns output of the final convolution layer

        Note that predefined arches downsample by a factor of about 32x

        Example:
            >>> # xdoctest: +REQUIRES(module:ndsampler)
            >>> from netharn.models.efficientnet import *  # NOQA
            >>> self = EfficientNet.from_name('efficientnet-b0')
            >>> self = self.train(False)
            >>> inputs = torch.rand(1, 3, 32, 32)
            >>> x = self.extract_features(inputs)
        """
        # Stem
        x = self._conv_stem(inputs)
        x = self._noli(self._bn0(x))

        # Blocks
        for idx, block in enumerate(self._blocks):
            drop_connect_rate = self._global_params.drop_connect_rate
            if drop_connect_rate:
                drop_connect_rate *= float(idx) / len(self._blocks)
            x = block(x, drop_connect_rate=drop_connect_rate)

        # Head
        x = self._noli(self._bn1(self._conv_head(x)))

        return x

    def _analytic_forward(self, inputs, _OutputFor, _Output, _Hidden,
                          **kwargs):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:ndsampler)
            >>> import netharn as nh
            >>> from netharn.models.efficientnet import *  # NOQA
            >>> self = EfficientNet.from_name('efficientnet-b0')
            >>> kwargs = self._analytic_shape_kw()
            >>> globals().update(kwargs)
            >>> inputs = (1, 3, 224, 224)
            >>> inputs = (1, 3, 32, 32)
            >>> outputs = self.output_shape_for(inputs)
            >>> import ubelt as ub
            >>> print(nh.util.align(ub.repr2(outputs.hidden.shallow(1), nl=-1), ':'))

            >>> print(nh.util.align(ub.repr2(outputs.hidden['block_0'].shallow(2), nl=-1), ':'))
            >>> print(nh.util.align(ub.repr2(outputs.hidden['block_1'].shallow(2), nl=-1), ':'))
            >>> print(nh.util.align(ub.repr2(outputs.hidden['block_2'].shallow(2), nl=-1), ':'))
            >>> print(nh.util.align(ub.repr2(outputs.hidden['block_3'].shallow(2), nl=-1), ':'))
            >>> print(nh.util.align(ub.repr2(outputs.hidden['block_14'].shallow(2), nl=-1), ':'))
            >>> print(nh.util.align(ub.repr2(outputs.hidden['block_15'].shallow(2), nl=-1), ':'))

            >>> self = EfficientNet.from_name('efficientnet-b7')
            >>> print('self.image_size = {!r}'.format(self.image_size))
            >>> self = EfficientNet.from_name('efficientnet-b6')
            >>> print('self.image_size = {!r}'.format(self.image_size))
            >>> self = EfficientNet.from_name('efficientnet-b3')
            >>> print('self.image_size = {!r}'.format(self.image_size))
            >>> self = EfficientNet.from_name('efficientnet-b2')
            >>> print('self.image_size = {!r}'.format(self.image_size))
            >>> self = EfficientNet.from_name('efficientnet-b1')
            >>> print('self.image_size = {!r}'.format(self.image_size))
            >>> self = EfficientNet.from_name('efficientnet-b0')
            >>> print('self.image_size = {!r}'.format(self.image_size))

            >>> inputs = (1, 3, 224, 224)
            >>> self = EfficientNet.from_name('efficientnet-b7')
            >>> print('self.image_size = {!r}'.format(self.image_size))
            >>> outputs = self.output_shape_for(inputs)
            >>> print(nh.util.align(ub.repr2(outputs.hidden.shallow(1), nl=-1), ':'))

            for name, layer in nh.util.trainable_layers(self, names=1):
                if hasattr(layer, 'image_size'):
                    print('name = {!r}'.format(name))
                    print('layer = {!r}'.format(layer))
                    print('layer.image_size = {!r}'.format(layer.image_size))

            >>> inputs = (1, 3, 224, 224)
            >>> self = EfficientNet.from_name('efficientnet-b0')
            >>> outputs = self.output_shape_for(inputs)
            >>> print(nh.util.align(ub.repr2(outputs.hidden.shallow(1), nl=-1), ':'))

            for name, layer in nh.util.trainable_layers(self, names=1):
                if hasattr(layer, 'image_size'):
                    print('name = {!r}'.format(name))
                    print('layer = {!r}'.format(layer))
                    print('layer.image_size = {!r}'.format(layer.image_size))
        """
        hidden = _Hidden()

        # NEEDS MORE BACKEND WORK

        bs = _OutputFor.shape(inputs)[0]

        x = inputs
        x = hidden['_conv_stem'] = _OutputFor(self._conv_stem)(x)
        x = hidden['_noli1'] = _OutputFor(self._noli)(x)

        for idx, block in enumerate(self._blocks):
            drop_connect_rate = self._global_params.drop_connect_rate
            if drop_connect_rate:
                drop_connect_rate *= float(idx) / len(self._blocks)
            x = hidden['block_{}'.format(idx)] = _OutputFor(block)(
                x, drop_connect_rate=drop_connect_rate)

        x = hidden['_noli2'] = _OutputFor(self._noli)(x)

        # Pooling and final linear layer
        x = _OutputFor(self._avg_pooling)(x)
        x = _OutputFor.view(x, bs, -1)
        x = _OutputFor(self._dropout)(x)
        x = _OutputFor(self._fc)(x)
        outputs = _Output.coerce(x, hidden)
        return outputs

    def forward(self, inputs):
        """
        Calls extract_features to extract features,
        applies final linear layer, and returns logits.
        """
        bs = inputs.size(0)
        # Convolution layers
        x = self.extract_features(inputs)

        # Pooling and final linear layer
        x = self._avg_pooling(x)
        x = x.view(bs, -1)
        x = self._dropout(x)
        x = self._fc(x)
        return x

    # TODO: Analytic forward

    @classmethod
    def from_params(cls, width, depth, size, dropout, **override_params):
        # note: all models have drop connect rate = 0.2
        blocks_args, global_params = Details.build_efficientnet_params(
            width_coefficient=width, depth_coefficient=depth,
            dropout_rate=dropout, image_size=size)
        global_params = global_params._replace(**override_params)

    @classmethod
    def from_name(EfficientNet, model_name, override_params=None):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:ndsampler)
            >>> model_name = 'efficientnet-b0'
            >>> override_params = None
        """
        Details._check_model_name_is_valid(model_name)
        blocks_args, global_params = Details._get_model_params(model_name, override_params)
        self = EfficientNet(blocks_args, global_params)
        self.model_name = model_name
        return self

    @classmethod
    def from_pretrained(EfficientNet, model_name, advprop=False,
                        override_params=None, in_channels=3):
        """
        Initialize the model from a pretrained state

        Example:
            >>> # xdoctest: +REQUIRES(--download)
            >>> # xdoctest: +REQUIRES(module:ndsampler)
            >>> from netharn.models.efficientnet import *  # NOQA
            >>> model = EfficientNet.from_pretrained('efficientnet-b0')
            >>> inputs = torch.rand(1, 3, 224, 224)
            >>> outputs = model.forward(inputs)

            >>> from netharn.models.efficientnet import *  # NOQA
            >>> model = EfficientNet.from_pretrained('efficientnet-b0', override_params={'noli': 'mish'}, advprop=True)
            >>> model = EfficientNet.from_pretrained('efficientnet-b1', override_params={'noli': 'mish'}, advprop=True)
            >>> model = EfficientNet.from_pretrained('efficientnet-b2', override_params={'noli': 'mish'}, advprop=True)
            >>> model = EfficientNet.from_pretrained('efficientnet-b3', override_params={'noli': 'mish'}, advprop=True)
            >>> model = EfficientNet.from_pretrained('efficientnet-b4', override_params={'noli': 'mish'}, advprop=True)
            >>> model = EfficientNet.from_pretrained('efficientnet-b5', override_params={'noli': 'mish'}, advprop=True)
            >>> model = EfficientNet.from_pretrained('efficientnet-b6', override_params={'noli': 'mish'}, advprop=True)
            >>> model = EfficientNet.from_pretrained('efficientnet-b7', override_params={'noli': 'mish'}, advprop=True)
        """
        if override_params is None:
            override_params = {}
        self = EfficientNet.from_name(model_name, override_params=override_params)
        num_classes = len(self.classes)
        Details.load_pretrained_weights(
            self, model_name, load_fc=(num_classes == 1000), advprop=advprop)
        if in_channels != 3:
            Conv2d = Conv2dDynamicSamePadding.forsize(image_size=self._global_params.image_size)
            out_channels = self.round_filters(32)
            self._conv_stem = Conv2d(in_channels, out_channels, kernel_size=3, stride=2, bias=False)
        return self

    def get_image_size(self):
        if self.model_name is not None:
            return Details.get_image_size(self.model_name)
        else:
            raise NotImplementedError('only know size for predefined models')


class Details(object):
    """
    Definition of models

    Helpers functions for loading model params
    """

    # Parameters for the entire model (stem, all blocks, and head)
    # url_map = {
    #     'efficientnet-b0': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b0-355c32eb.pth',
    #     'efficientnet-b1': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b1-f1951068.pth',
    #     'efficientnet-b2': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b2-8bb594d6.pth',
    #     'efficientnet-b3': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b3-5fb5a3c3.pth',
    #     'efficientnet-b4': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b4-6ed6700e.pth',
    #     'efficientnet-b5': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b5-b6417697.pth',

    #     # 'efficientnet-b1': 'https://www.dropbox.com/s/6745ear79b1ltkh/efficientnet-b1-ef6aa7.pth?dl=1',
    #     # 'efficientnet-b2': 'https://www.dropbox.com/s/0dhtv1t5wkjg0iy/efficientnet-b2-7c98aa.pth?dl=1',
    #     # 'efficientnet-b3': 'https://www.dropbox.com/s/5uqok5gd33fom5p/efficientnet-b3-bdc7f4.pth?dl=1',
    #     # 'efficientnet-b4': 'https://www.dropbox.com/s/y2nqt750lixs8kc/efficientnet-b4-3e4967.pth?dl=1',
    #     # 'efficientnet-b5': 'https://www.dropbox.com/s/qxonlu3q02v9i47/efficientnet-b5-4c7978.pth?dl=1',

    #     'efficientnet-b6': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b6-c76e70fd.pth',
    #     'efficientnet-b7': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b7-dcc49843.pth',
    # }

    # url_map_advprop = {
    #     'efficientnet-b0': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b0-b64d5a18.pth',
    #     'efficientnet-b1': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b1-0f3ce85a.pth',
    #     'efficientnet-b2': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b2-6e9d97e5.pth',
    #     'efficientnet-b3': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b3-cdd7c0f4.pth',
    #     'efficientnet-b4': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b4-44fb3a87.pth',
    #     'efficientnet-b5': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b5-86493f6b.pth',
    #     'efficientnet-b6': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b6-ac80338e.pth',
    #     'efficientnet-b7': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b7-4652b6dd.pth',
    #     'efficientnet-b8': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b8-22a8fe65.pth',
    # }

    url_map = {
        'efficientnet-b0': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b0-355c32eb.pth',
        'efficientnet-b1': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b1-f1951068.pth',
        'efficientnet-b2': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b2-8bb594d6.pth',
        'efficientnet-b3': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b3-5fb5a3c3.pth',
        'efficientnet-b4': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b4-6ed6700e.pth',
        'efficientnet-b5': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b5-b6417697.pth',
        'efficientnet-b6': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b6-c76e70fd.pth',
        'efficientnet-b7': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b7-dcc49843.pth',
    }

    url_map_advprop = {
        'efficientnet-b0': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b0-b64d5a18.pth',
        'efficientnet-b1': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b1-0f3ce85a.pth',
        'efficientnet-b2': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b2-6e9d97e5.pth',
        'efficientnet-b3': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b3-cdd7c0f4.pth',
        'efficientnet-b4': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b4-44fb3a87.pth',
        'efficientnet-b5': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b5-86493f6b.pth',
        'efficientnet-b6': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b6-ac80338e.pth',
        'efficientnet-b7': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b7-4652b6dd.pth',
        'efficientnet-b8': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b8-22a8fe65.pth',
    }

    @classmethod
    def get_image_size(cls, model_name):
        Details._check_model_name_is_valid(model_name)
        _, _, res, _ = Details._named_params(model_name)
        return res

    @classmethod
    def _check_model_name_is_valid(cls, model_name):
        """ Validates model name. """
        valid_models = ['efficientnet-b{}'.format(i) for i in range(9)]
        if model_name not in valid_models:
            raise ValueError('model_name should be one of: ' + ', '.join(valid_models))

    @classmethod
    def _get_model_params(cls, model_name, override_params=None):
        """
        Get the block args and global params for a given model

        Example:
            Details._get_model_params('efficientnet-b0')
        """
        if model_name.startswith('efficientnet'):
            w, d, s, p = Details._named_params(model_name)
            # note: all models have drop connect rate = 0.2
            blocks_args, global_params = Details.build_efficientnet_params(
                width_coefficient=w, depth_coefficient=d, dropout_rate=p, image_size=s)
        else:
            raise NotImplementedError('model name is not pre-defined: %s' % model_name)
        if override_params:
            # ValueError will be raised here if override_params has fields not included in global_params.
            global_params = global_params._replace(**override_params)
        return blocks_args, global_params

    @classmethod
    def _named_params(cls, model_name):
        """ Map EfficientNet model name to parameter coefficients. """
        params_dict = {
            # Coefficients:   width,depth,res,dropout
            'efficientnet-b0': (1.0, 1.0, 224, 0.2),
            'efficientnet-b1': (1.0, 1.1, 240, 0.2),
            'efficientnet-b2': (1.1, 1.2, 260, 0.3),
            'efficientnet-b3': (1.2, 1.4, 300, 0.3),
            'efficientnet-b4': (1.4, 1.8, 380, 0.4),
            'efficientnet-b5': (1.6, 2.2, 456, 0.4),
            'efficientnet-b6': (1.8, 2.6, 528, 0.5),
            'efficientnet-b7': (2.0, 3.1, 600, 0.5),
            'efficientnet-b8': (2.2, 3.6, 672, 0.5),
            'efficientnet-l2': (4.3, 5.3, 800, 0.5),
        }
        return params_dict[model_name]

    @staticmethod
    def decode(string_list):
        """
        Decodes a list of string notations to specify blocks inside the network.
        :param string_list: a list of strings, each string is a notation of block
        :return: a list of BlockArgs namedtuples of block args
        """
        assert isinstance(string_list, list)
        blocks_args = []
        for block_string in string_list:
            blocks_args.append(Details._decode_block_string(block_string))
        return blocks_args

    @staticmethod
    def _decode_block_string(block_string):
        """ Gets a block through a string notation of arguments. """
        assert isinstance(block_string, str)

        ops = block_string.split('_')
        options = {}
        for op in ops:
            splits = re.split(r'(\d.*)', op)
            if len(splits) >= 2:
                key, value = splits[:2]
                options[key] = value

        # Check stride
        assert (('s' in options and len(options['s']) == 1) or
                (len(options['s']) == 2 and options['s'][0] == options['s'][1]))

        return Details.BlockArgs(
            kernel_size=int(options['k']),
            num_repeat=int(options['r']),
            input_filters=int(options['i']),
            output_filters=int(options['o']),
            expand_ratio=int(options['e']),
            id_skip=('noskip' not in block_string),
            se_ratio=float(options['se']) if 'se' in options else None,
            stride=[int(options['s'][0])])

    # Parameters for an individual model block
    BlockArgs = collections.namedtuple('BlockArgs', [
        'kernel_size', 'num_repeat', 'input_filters', 'output_filters',
        'expand_ratio', 'id_skip', 'stride', 'se_ratio'])
    BlockArgs.__new__.__defaults__ = (None,) * len(BlockArgs._fields)

    @staticmethod
    def encode(blocks_args):
        """
        Encodes a list of BlockArgs to a list of strings.
        :param blocks_args: a list of BlockArgs namedtuples of block args
        :return: a list of strings, each string is a notation of block
        """
        block_strings = []
        for block in blocks_args:
            block_strings.append(Details._encode_block_string(block))
        return block_strings

    @staticmethod
    def _encode_block_string(block):
        """Encodes a block to a string."""
        args = [
            'r%d' % block.num_repeat,
            'k%d' % block.kernel_size,
            's%d%d' % (block.strides[0], block.strides[1]),
            'e%s' % block.expand_ratio,
            'i%d' % block.input_filters,
            'o%d' % block.output_filters
        ]
        if 0 < block.se_ratio <= 1:
            args.append('se%s' % block.se_ratio)
        if block.id_skip is False:
            args.append('noskip')
        return '_'.join(args)

    GlobalParams = collections.namedtuple('GlobalParams', [
        'batch_norm_momentum', 'batch_norm_epsilon', 'dropout_rate',
        'num_classes', 'width_coefficient', 'depth_coefficient',
        'depth_divisor', 'min_depth', 'drop_connect_rate', 'image_size',
        'classes', 'noli'])

    # Change namedtuple defaults
    GlobalParams.__new__.__defaults__ = (None,) * len(GlobalParams._fields)

    @staticmethod
    def build_efficientnet_params(width_coefficient=None,
                                  depth_coefficient=None, dropout_rate=0.2,
                                  drop_connect_rate=0.2, image_size=None,
                                  num_classes=1000):
        """
        Creates a efficientnet parameters

        Example:
            Details.build_efficientnet_params(None, None, image_size=512)
        """

        blocks_args = [
            'r1_k3_s11_e1_i32_o16_se0.25', 'r2_k3_s22_e6_i16_o24_se0.25',
            'r2_k5_s22_e6_i24_o40_se0.25', 'r3_k3_s22_e6_i40_o80_se0.25',
            'r3_k5_s11_e6_i80_o112_se0.25', 'r4_k5_s22_e6_i112_o192_se0.25',
            'r1_k3_s11_e6_i192_o320_se0.25',
        ]
        blocks_args = Details.decode(blocks_args)

        global_params = Details.GlobalParams(
            batch_norm_momentum=0.99,
            batch_norm_epsilon=1e-3,
            dropout_rate=dropout_rate,
            drop_connect_rate=drop_connect_rate,
            num_classes=num_classes,
            width_coefficient=width_coefficient,
            depth_coefficient=depth_coefficient,
            depth_divisor=8,
            min_depth=None,
            image_size=image_size,
            classes=None,
            noli='swish'
        )
        return blocks_args, global_params

    @staticmethod
    def load_pretrained_weights(model, model_name, load_fc=True, advprop=False):
        """ Loads pretrained weights, and downloads if loading for the first time. """
        # AutoAugment or Advprop (different preprocessing)
        url_map_ = Details.url_map_advprop if advprop else Details.url_map
        state_dict = model_zoo.load_url(url_map_[model_name])
        if load_fc:
            model.load_state_dict(state_dict)
        else:
            state_dict.pop('_fc.weight')
            state_dict.pop('_fc.bias')
            res = model.load_state_dict(state_dict, strict=False)
            assert set(res.missing_keys) == set(['_fc.weight', '_fc.bias']), (
                'issue loading pretrained weights')
        print('Loaded pretrained weights for {}'.format(model_name))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import copy
import json
import six
import torch
import numpy as np
import ubelt as ub
from collections import OrderedDict


def walk_json(node):
    for key, val in node.items():
        if isinstance(val, dict):
            for item in walk_json(val):
                yield item
        elif isinstance(val, list):
            for subval in val:
                if isinstance(subval, dict):
                    for item in walk_json(subval):
                        yield item
                else:
                    assert False
        else:
            yield key, val


class LossyJSONEncoder(json.JSONEncoder):
    """
    Helps cooerce objects into a json-serializable format. Note that this is a
    lossy process. Information about object types / array types are lost. Only
    info directly translatable to json primitives are preserved as those
    primitive types.  (e.g: tuples and ndarrays are encoded as lists, and
    objects only remember their dict of attributes).

    Example:
        >>> import json
        >>> class MyClass(object):
        >>>     def __init__(self, foo='bar'):
        >>>         self.foo = foo
        >>>         self.spam = 32
        >>>         self.eggs = np.array([32])
        >>>     def __json__(self):
        >>>         return {self.__class__.__name__: self.__dict__}
        >>> self = MyClass()
        >>> text = json.dumps(self, cls=LossyJSONEncoder)
        >>> reloaded = json.loads(text)
        >>> expected = {'MyClass': {'foo': 'bar', 'spam': 32, 'eggs': [32]}}
        >>> assert reloaded == expected
    """
    def default(self, obj):
        if hasattr(obj, '__json__'):
            return obj.__json__()
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.generic):
            return obj.item()
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(LossyJSONEncoder, self).default(obj)
        # return json.JSONEncoder.default(self, obj)


class NumpyEncoder(json.JSONEncoder):
    """
    https://stackoverflow.com/questions/3488934/simplejson-and-numpy-array
    """

    def default(self, obj):
        """If input object is an ndarray it will be converted into a dict
        holding dtype, shape and the data, base64 encoded.
        """
        import base64
        if isinstance(obj, np.ndarray):
            if obj.flags['C_CONTIGUOUS']:
                obj_data = obj.data
            else:
                cont_obj = np.ascontiguousarray(obj)
                assert(cont_obj.flags['C_CONTIGUOUS'])
                obj_data = cont_obj.data
            data_b64 = base64.b64encode(obj_data)
            return dict(__ndarray__=data_b64,
                        dtype=str(obj.dtype),
                        shape=obj.shape)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder(self, obj)

    @staticmethod
    def json_numpy_obj_hook(dct):
        """Decodes a previously encoded numpy ndarray with proper shape and dtype.

        :param dct: (dict) json encoded ndarray
        :return: (ndarray) if input was an encoded ndarray
        """
        import base64
        if isinstance(dct, dict) and '__ndarray__' in dct:
            data = base64.b64decode(dct['__ndarray__'])
            return np.frombuffer(data, dct['dtype']).reshape(dct['shape'])
        return dct


def write_json(fpath, data):
    """
    Write human readable json files
    """
    try:
        import pandas as pd
    except ImportError:
        pd = None

    if pd and isinstance(data, pd.DataFrame):
        # pretty pandas
        json_text = (json.dumps(json.loads(data.to_json()), indent=4))
    elif isinstance(data, dict):
        json_text = json.dumps(data, cls=LossyJSONEncoder, indent=4)
    else:
        raise TypeError(type(data))
    ub.writeto(fpath, json_text)


def read_json(fpath):
    """
    Write human readable json files
    """
    if isinstance(fpath, six.string_types):
        return json.load(open(fpath, 'r'))
    else:
        return json.load(fpath)


def ensure_json_serializable(dict_, normalize_containers=False, verbose=0):
    """
    Attempt to convert common types (e.g. numpy) into something json complient

    Convert numpy and tuples into lists

    Args:
        normalize_containers (bool, default=False):
            if True, normalizes dict containers to be standard python
            structures.

    Example:
        >>> from netharn.hyperparams import *  # NOQA
        >>> data = ub.ddict(lambda: int)
        >>> data['foo'] = ub.ddict(lambda: int)
        >>> data['bar'] = np.array([1, 2, 3])
        >>> data['foo']['a'] = 1
        >>> data['foo']['b'] = torch.FloatTensor([1, 2, 3])
        >>> result = ensure_json_serializable(data, normalize_containers=True)
        >>> assert type(result) is dict
    """
    dict_ = copy.deepcopy(dict_)

    def _norm_container(c):
        if isinstance(c, dict):
            # Cast to a normal dictionary
            if isinstance(c, OrderedDict):
                if type(c) is not OrderedDict:
                    c = OrderedDict(c)
            else:
                if type(c) is not dict:
                    c = dict(c)
        return c

    # inplace convert any ndarrays to lists
    def _walk_json(data, prefix=[]):
        items = None
        if isinstance(data, list):
            items = enumerate(data)
        elif isinstance(data, tuple):
            items = enumerate(data)
        elif isinstance(data, dict):
            items = data.items()
        else:
            raise TypeError(type(data))

        root = prefix
        level = {}
        for key, value in items:
            level[key] = value

        # yield a dict so the user can choose to not walk down a path
        yield root, level

        for key, value in level.items():
            if isinstance(value, (dict, list, tuple)):
                path = prefix + [key]
                for _ in _walk_json(value, prefix=path):
                    yield _

    def _convert(dict_, root, key, new_value):
        d = dict_
        for k in root:
            d = d[k]
        d[key] = new_value

    def _flatmap(func, data):
        if isinstance(data, list):
            return [_flatmap(func, item) for item in data]
        else:
            return func(data)

    to_convert = []
    for root, level in ub.ProgIter(_walk_json(dict_), desc='walk json',
                                   verbose=verbose):
        for key, value in level.items():
            if isinstance(value, tuple):
                # Convert tuples on the fly so they become mutable
                new_value = list(value)
                _convert(dict_, root, key, new_value)
            elif isinstance(value, np.ndarray):
                new_value = value.tolist()
                if 0:
                    if len(value.shape) == 1:
                        if value.dtype.kind in {'i', 'u'}:
                            new_value = list(map(int, new_value))
                        elif value.dtype.kind in {'f'}:
                            new_value = list(map(float, new_value))
                        elif value.dtype.kind in {'c'}:
                            new_value = list(map(complex, new_value))
                        else:
                            pass
                    else:
                        if value.dtype.kind in {'i', 'u'}:
                            new_value = _flatmap(int, new_value)
                        elif value.dtype.kind in {'f'}:
                            new_value = _flatmap(float, new_value)
                        elif value.dtype.kind in {'c'}:
                            new_value = _flatmap(complex, new_value)
                        else:
                            pass
                            # raise TypeError(value.dtype)
                to_convert.append((root, key, new_value))
            elif isinstance(value, torch.Tensor):
                new_value = value.data.cpu().numpy().tolist()
                to_convert.append((root, key, new_value))
            elif isinstance(value, (np.int16, np.int32, np.int64,
                                    np.uint16, np.uint32, np.uint64)):
                new_value = int(value)
                to_convert.append((root, key, new_value))
            elif isinstance(value, (np.float32, np.float64)):
                new_value = float(value)
                to_convert.append((root, key, new_value))
            elif isinstance(value, (np.complex64, np.complex128)):
                new_value = complex(value)
                to_convert.append((root, key, new_value))
            elif hasattr(value, '__json__'):
                new_value = value.__json__()
                to_convert.append((root, key, new_value))
            elif normalize_containers:
                if isinstance(value, dict):
                    new_value = _norm_container(value)
                    to_convert.append((root, key, new_value))

    for root, key, new_value in to_convert:
        _convert(dict_, root, key, new_value)

    if normalize_containers:
        # normalize the outer layer
        dict_ = _norm_container(dict_)
    return dict_



import torch.utils.data.sampler as torch_sampler
import torch


class MultiScaleBatchSampler(torch_sampler.BatchSampler):
    """
    Indicies returned in the batch are tuples indicating data index and scale
    index. Requires that dataset has a `multi_scale_inp_size` attribute.

    Args:
        sampler (Sampler): Base sampler. Must have a data_source attribute.
        batch_size (int): Size of mini-batch.
        drop_last (bool): If ``True``, the sampler will drop the last batch if
            its size would be less than ``batch_size``
        resample_freq (int): how often to change scales. if None, then
            only one scale is used.

    Example:
        >>> import torch.utils.data as torch_data
        >>> class DummyDatset(torch_data.Dataset):
        >>>     def __init__(self):
        >>>         super(DummyDatset, self).__init__()
        >>>         self.multi_scale_inp_size = [1, 2, 3, 4]
        >>>     def __len__(self):
        >>>         return 34
        >>> batch_size = 16
        >>> data_source = DummyDatset()
        >>> sampler1 = torch_sampler.RandomSampler(data_source)
        >>> sampler2 = torch_sampler.SequentialSampler(data_source)
        >>> rand = MultiScaleBatchSampler(sampler1, resample_freq=10)
        >>> seq = MultiScaleBatchSampler(sampler2, resample_freq=None)
        >>> rand_idxs = list(iter(rand))
        >>> seq_idxs = list(iter(seq))
        >>> assert len(rand_idxs[0]) == 16
        >>> assert len(rand_idxs[0][0]) == 2
        >>> assert len(rand_idxs[-1]) == 2
        >>> assert {len({x[1] for x in xs}) for xs in rand_idxs} == {1}
        >>> assert {x[1] for xs in seq_idxs for x in xs} == {None}
    """

    def __init__(self, sampler, batch_size=16, drop_last=False,
                 resample_freq=10):
        self.sampler = sampler
        self.batch_size = batch_size
        self.drop_last = drop_last
        self.num_scales = len(sampler.data_source.multi_scale_inp_size)
        self.resample_freq = resample_freq

    def __iter__(self):
        batch = []
        if self.resample_freq:
            scale_index = int(torch.rand(1) * self.num_scales)
        else:
            scale_index = None

        for idx in self.sampler:
            batch.append((int(idx), scale_index))
            if len(batch) == self.batch_size:
                yield batch
                if self.resample_freq and idx % self.resample_freq == 0:
                    # choose a new scale index every `resample_freq` batches
                    scale_index = int(torch.rand(1) * self.num_scales)
                batch = []
        if len(batch) > 0 and not self.drop_last:
            yield batch


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m netharn.models.yolo2.multiscale_batch_sampler all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# -*- coding: utf-8 -*-
"""
NOTE: THIS FILE IS DEPRECATED

Processing for filenames. The logic is relatively hacky.

pip install pygtrie
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from os.path import commonprefix, isdir, dirname
import numpy as np  # NOQA


def shortest_unique_prefixes(items, sep=None, allow_simple=True, min_length=0, allow_end=False):
    r"""
    The shortest unique prefix algorithm.

    Args:
        items (list of str): returned prefixes will be unique wrt this set
        sep (str): if specified, all characters between separators are treated
            as a single symbol. Makes the algo much faster.
        allow_simple (bool): if True tries to construct a simple feasible
            solution before resorting to the optimal trie algorithm.
        min_length (int): minimum length each prefix can be
        allow_end (bool): if True allows for string terminators to be
            considered in the prefix

    Returns:
        list of str: a prefix for each item that uniquely identifies it
           wrt to the original items.

    References:
        http://www.geeksforgeeks.org/find-all-shortest-unique-prefixes-to-represent-each-word-in-a-given-list/
        https://github.com/Briaares/InterviewBit/blob/master/Level6/Shortest%20Unique%20Prefix.cpp

    Requires:
        pip install pygtrie

    Example:
        >>> # xdoctest: +REQUIRES(--pygtrie)
        >>> items = ["zebra", "dog", "duck", "dove"]
        >>> shortest_unique_prefixes(items)
        ['z', 'dog', 'du', 'dov']

    Timeing:
        >>> # DISABLE_DOCTEST
        >>> # make numbers larger to stress test
        >>> # L = max length of a string, N = number of strings,
        >>> # C = smallest gaurenteed common length
        >>> # (the setting N=10000, L=100, C=20 is feasible we are good)
        >>> import ubelt as ub
        >>> import random
        >>> def make_data(N, L, C):
        >>>     rng = random.Random(0)
        >>>     return [''.join(['a' if i < C else chr(rng.randint(97, 122))
        >>>                      for i in range(L)]) for _ in range(N)]
        >>> items = make_data(N=1000, L=10, C=0)
        >>> ub.Timerit(3).call(shortest_unique_prefixes, items).print()
        Timed for: 3 loops, best of 3
            time per loop: best=24.54 ms, mean=24.54 ± 0.0 ms
        >>> items = make_data(N=1000, L=100, C=0)
        >>> ub.Timerit(3).call(shortest_unique_prefixes, items).print()
        Timed for: 3 loops, best of 3
            time per loop: best=155.4 ms, mean=155.4 ± 0.0 ms
        >>> items = make_data(N=1000, L=100, C=70)
        >>> ub.Timerit(3).call(shortest_unique_prefixes, items).print()
        Timed for: 3 loops, best of 3
            time per loop: best=232.8 ms, mean=232.8 ± 0.0 ms
        >>> items = make_data(N=10000, L=250, C=20)
        >>> ub.Timerit(3).call(shortest_unique_prefixes, items).print()
        Timed for: 3 loops, best of 3
            time per loop: best=4.063 s, mean=4.063 ± 0.0 s
    """
    import pygtrie
    if len(set(items)) != len(items):
        raise ValueError('inputs must be unique')

    # construct trie
    if sep is None:
        trie = pygtrie.CharTrie.fromkeys(items, value=0)
    else:
        # In some simple cases we can avoid constructing a trie
        if allow_simple:
            tokens = [item.split(sep) for item in items]
            simple_solution = [t[0] for t in tokens]
            if len(simple_solution) == len(set(simple_solution)):
                return simple_solution
            for i in range(2, 10):
                # print('return simple solution at i = {!r}'.format(i))
                simple_solution = ['-'.join(t[:i]) for t in tokens]
                if len(simple_solution) == len(set(simple_solution)):
                    return simple_solution

        trie = pygtrie.StringTrie.fromkeys(items, value=0, separator=sep)

    # Set the value (frequency) of all nodes to zero.
    for node in _trie_iternodes(trie):
        node.value = 0

    # For each item trace its path and increment frequencies
    for item in items:
        final_node, trace = trie._get_node(item)
        for key, node in trace:
            node.value += 1

    # if not isinstance(node.value, int):
    #     node.value = 0

    # Query for the first prefix with frequency 1 for each item.
    # This is the shortest unique prefix over all items.
    unique = []
    for item in items:
        freq = None
        for prefix, freq in trie.prefixes(item):
            if freq == 1 and len(prefix) >= min_length:
                break
        if not allow_end:
            assert freq == 1, 'item={} has no unique prefix. freq={}'.format(item, freq)
        # print('items = {!r}'.format(items))
        unique.append(prefix)
    return unique


def _trie_iternodes(self):
    """
    Generates all nodes in the trie

    # Hack into the internal structure and insert frequencies at each node
    """
    from collections import deque
    stack = deque([[self._root]])
    while stack:
        for node in stack.pop():
            yield node
            try:
                # only works in pygtrie-2.2 broken in pygtrie-2.3.2
                stack.append(node.children.values())
            except AttributeError:
                stack.append([v for k, v in node.children.iteritems()])


def shortest_unique_suffixes(items, sep=None):
    r"""
    CommandLine:
        xdoctest -m netharn.util.util_fname shortest_unique_suffixes

    Example:
        >>> # xdoctest: +REQUIRES(--pygtrie)
        >>> items = ["zebra", "dog", "duck", "dove"]
        >>> shortest_unique_suffixes(items)
        ['a', 'g', 'k', 'e']

    Example:
        >>> # xdoctest: +REQUIRES(--pygtrie)
        >>> items = ["aa/bb/cc", "aa/bb/bc", "aa/bb/dc", "aa/cc/cc"]
        >>> shortest_unique_suffixes(items)
    """
    snoitpo = [p[::-1] for p in items]
    sexiffus = shortest_unique_prefixes(snoitpo, sep=sep)
    suffixes = [s[::-1] for s in sexiffus]
    return suffixes


def dumpsafe(paths, repl='<sl>'):
    """
    enforces that filenames will not conflict.
    Removes common the common prefix, and replaces slashes with <sl>

    Ignore:
        >>> # xdoctest: +REQUIRES(--pygtrie)
        >>> paths = ['foo/{:04d}/{:04d}'.format(i, j) for i in range(2) for j in range(20)]
        >>> list(dumpsafe(paths, '-'))
    """
    common_pref = commonprefix(paths)
    if not isdir(common_pref):
        im_pref = dirname(common_pref)
        if common_pref[len(im_pref):len(im_pref) + 1] == '/':
            im_pref += '/'
        elif common_pref[len(im_pref):len(im_pref) + 1] == '\\':
            im_pref += '\\'
    else:
        im_pref = common_pref

    start = len(im_pref)
    dump_paths = (
        p[start:].replace('/', repl).replace('\\', repl)  # faster
        # relpath(p, im_pref).replace('/', repl).replace('\\', repl)
        for p in paths
    )
    return dump_paths


def _fast_name_we(fname):
    # Assume that extensions are no more than 7 chars for extra speed
    pos = fname.rfind('.', -7)
    return fname if pos == -1 else fname[:pos]


def _fast_basename_we(fname):
    slashpos = fname.rfind('/')
    base = fname if slashpos == -1 else fname[slashpos + 1:]
    pos = base.rfind('.', -slashpos)
    base_we = base if pos == -1 else base[:pos]
    return base_we


def _safepaths(paths):
    r"""
    Ignore:
        x = '/home/local/KHQ/jon.crall/code/netharn/netharn/live/urban_train.py'
        import re
        %timeit splitext(x.replace('<sl>', '-').replace('_', '-'))[0]
        %timeit splitext(re.sub('<sl>|_', '-', x))
        %timeit x[:x.rfind('.')].replace('<sl>', '-').replace('_', '-')
        %timeit _fast_name_we(x)
        %timeit x[:x.rfind('.')]

        >>> paths = ['foo/{:04d}/{:04d}'.format(i, j) for i in range(2) for j in range(20)]
        >>> _safepaths(paths)
    """
    safe_paths = [
        # faster than splitext
        _fast_name_we(x).replace('_', '-').replace('<sl>', '-')
        # splitext(x.replace('<sl>', '-').replace('_', '-'))[0]
        for x in dumpsafe(paths, repl='-')
    ]
    return safe_paths


def align_paths(paths1, paths2):
    r"""
    return path2 in the order of path1

    This function will only work where file types (i.e. image / groundtruth)
    are specified by EITHER a path prefix XOR a path suffix (note this is an
    exclusive or. do not mix prefixes and suffixes), either as part of a
    filename or parent directory. In the case of a filename it is assumped this
    "type identifier" is separated from the rest of the path by an underscore
    or hyphen.

    paths1, paths2 = gt_paths, pred_paths

    Doctest:
        >>> # xdoc: +REQUIRES(--pygtrie)
        >>> def test_gt_arrangements(paths1, paths2, paths2_):
        >>>     # no matter what order paths2_ comes in, it should align with the groundtruth
        >>>     assert align_paths(paths1, paths2_) == paths2
        >>>     assert align_paths(paths1[::-1], paths2_) == paths2[::-1]
        >>>     assert align_paths(paths1[0::2] + paths1[1::2], paths2_) == paths2[0::2] + paths2[1::2]
        >>>     sortx = np.arange(len(paths1))
        >>>     np.random.shuffle(sortx)
        >>>     assert align_paths(list(np.take(paths1, sortx)), paths2_) == list(np.take(paths2, sortx))
        >>> #
        >>> def test_input_arrangements(paths1, paths2):
        >>>     paths2_ = list(paths2)
        >>>     test_gt_arrangements(paths1, paths2, paths2_)
        >>>     test_gt_arrangements(paths1, paths2, paths2_[::-1])
        >>>     np.random.shuffle(paths2_)
        >>>     test_gt_arrangements(paths1, paths2, paths2_)
        >>> paths1 = ['foo/{:04d}/{:04d}'.format(i, j) for i in range(2) for j in range(20)]
        >>> paths2 = ['bar/{:04d}/{:04d}'.format(i, j) for i in range(2) for j in range(20)]
        >>> test_input_arrangements(paths1, paths2)
        >>> paths1 = ['foo/{:04d}/{:04d}'.format(i, j) for i in range(2) for j in range(20)]
        >>> paths2 = ['bar<sl>{:04d}<sl>{:04d}'.format(i, j) for i in range(2) for j in range(20)]
        >>> test_input_arrangements(paths1, paths2)

    Speed:
        >>> import ubelt as ub
        >>> paths1 = [ub.expandpath('~/foo/{:04d}/{:04d}').format(i, j) for i in range(2) for j in range(10000)]
        >>> paths2 = [ub.expandpath('~/bar/{:04d}/{:04d}').format(i, j) for i in range(2) for j in range(10000)]
        >>> np.random.shuffle(paths2)
        >>> aligned = align_paths(paths1, paths2)

        items = [p[::-1] for p in _safepaths(paths1)]

    Ignore:
        >>> # pathological case (can we support this?)
        >>> aligned = [
        >>>     ('ims/aaa.png', 'gts/aaa.png'),
        >>>     ('ims/bbb.png', 'gts/bbb.png'),
        >>>     ('ims/ccc.png', 'gts/ccc.png'),
        >>>     # ---
        >>>     ('aaa/im.png', 'aaa/gt.png'),
        >>>     ('bbb/im.png', 'bbb/gt.png'),
        >>>     ('ccc/im.png', 'ccc/gt.png'),
        >>>     # ---
        >>>     ('ims/im-aaa.png', 'gts/gt-aaa.png'),
        >>>     ('ims/im-bbb.png', 'gts/gt-bbb.png'),
        >>>     ('ims/im-ccc.png', 'gts/gt-ccc.png'),
        >>>     # ---
        >>>     ('ims/aaa-im.png', 'gts/aaa-gt.png'),
        >>>     ('ims/bbb-im.png', 'gts/bbb-gt.png'),
        >>>     ('ims/ccc-im.png', 'gts/ccc-gt.png'),
        >>> ]
        >>> paths1, paths2 = zip(*aligned)

    """

    def comparable_unique_path_ids(paths1, paths2):
        """
        Given two unordered sets of paths (that are assumed to have some unknown
        correspondence) we find a unique id for each path in each set such that
        they can be aligned.
        """
        assert len(paths1) == len(paths2), (
            'cannot align unequal no of items {} != {}.'.format(len(paths1), len(paths2)))

        do_quick_check = True
        if do_quick_check:
            # First check the simple thing: do they have unique corresponding
            # basenames. If not we do something a bit more complex.
            simple_unique1 = list(map(_fast_basename_we, paths1))
            simple_unique_set1 = set(simple_unique1)
            if len(simple_unique_set1) == len(paths1):
                simple_unique2 = list(map(_fast_basename_we, paths2))
                simple_unique_set2 = set(simple_unique2)
                if simple_unique_set2 == simple_unique_set1:
                    return simple_unique1, simple_unique2

        safe_paths1 = _safepaths(paths1)
        safe_paths2 = _safepaths(paths2)

        # unique identifiers that should be comparable
        unique1 = shortest_unique_suffixes(safe_paths1, sep='-')
        unique2 = shortest_unique_suffixes(safe_paths2, sep='-')

        def not_comparable_msg():
            return '\n'.join([
                'paths are not comparable'
                'safe_paths1 = {}'.format(safe_paths1[0:3]),
                'safe_paths2 = {}'.format(safe_paths1[0:3]),
                'paths1 = {}'.format(safe_paths1[0:3]),
                'paths2 = {}'.format(safe_paths1[0:3]),
                'unique1 = {}'.format(unique1[0:3]),
                'unique2 = {}'.format(unique2[0:3]),
            ])

        try:
            # Assert these are unique identifiers common between paths
            assert sorted(set(unique1)) == sorted(unique1), not_comparable_msg()
            assert sorted(set(unique2)) == sorted(unique2), not_comparable_msg()
            assert sorted(unique1) == sorted(unique2), not_comparable_msg()
        except AssertionError:
            unique1 = shortest_unique_prefixes(safe_paths1, sep='-')
            unique2 = shortest_unique_prefixes(safe_paths2, sep='-')
            # Assert these are unique identifiers common between paths
            assert sorted(set(unique1)) == sorted(unique1), not_comparable_msg()
            assert sorted(set(unique2)) == sorted(unique2), not_comparable_msg()
            assert sorted(unique1) == sorted(unique2), not_comparable_msg()
        return unique1, unique2

    unique1, unique2 = comparable_unique_path_ids(paths1, paths2)

    lookup = {k: v for v, k in enumerate(unique1)}
    sortx = np.argsort([lookup[u] for u in unique2])

    sorted_paths2 = [paths2[x] for x in sortx]
    return sorted_paths2


def check_aligned(paths1, paths2):
    from os.path import basename
    if len(paths1) != len(paths2):
        return False

    # Try to short circuit common cases
    basenames1 = map(basename, paths1)
    basenames2 = map(basename, paths2)
    if all(p1 == p2 for p1, p2 in zip(basenames1, basenames2)):
        return True

    try:
        # Full case
        aligned_paths2 = align_paths(paths1, paths2)
    except AssertionError:
        return False
    return aligned_paths2 == paths2


if __name__ == '__main__':
    r"""
    CommandLine:
        xdoctest -m netharn.util.util_fname all --pygtrie
    """
    import xdoctest
    xdoctest.doctest_module(__file__)
